package VKRep;

import java.util.*;
import java.io.*;

/**
 * <p>Title: Base abstract class for all generator report by VK Reports</p>
 * <p>Description: Standard set of properties and methods for all generator reports.</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Vlad Karpov
 * @version 1.0
 */

public abstract class TReportAbstract implements Serializable, Runnable {

  /**
   *
   * <p>Title: The header of BIFF record.</p>
   * <p>Description: Any record in BIFF file (binary interchange file format)
   * is begin with standard header consits
   * of opcode and length of this record.</p>
   * <p>This class represent of this BIFF record header.</p>
   * <p>Copyright: Copyright (c) 2003</p>
   * <p>Company: </p>
   * @author Vlad Karpov
   * @version 1.0
   */
  protected class TBiffRecHeader {

    /**
     * Constructor.
     * Set to 0 of opcode and length.
     */
    public TBiffRecHeader() {
      opcode = 0;
      length = 0;
    }

    /**
     * Opcode field of BIFF file record header.
     */
    public short opcode;
    /**
     * Length record field of BIFF file record header.
     */
    public short length;

  }

  /**
   *
   * <p>Title: Class TBiffRec represents record of BIFF file itself.</p>
   * <p>Description: The object of this class holds BIFF file record</p>
   * <p>It conststs of BIFF record header and data of record.</p>
   * <p>Copyright: Copyright (c) 2003</p>
   * <p>Company: </p>
   * @author not attributable
   * @version 1.0
   */
  protected class TBiffRec {

    /**
     * Constructor.
     * Initialise object.
     */
    public TBiffRec() {
      head = new TBiffRecHeader();
      data = null;
      rb = 0;
      cb = 0;
      r = 0;
      c = 0;
      col_b = 0;
      col_e = 0;
      c1 = 0;
      c2 = 0;
      w = 0;
      d = null;
    }

    /**
     * Header of BIFF file record.
     * Object of class TBiffRecHeader.
     */
    public TBiffRecHeader head;
    /**
     * Data of record. The part of record from header
     * and length of head.length
     */
    public byte[] data;
    /**
     * Number of row relatively begin row section.
     */
    public int rb;
    /**
     * Number of collumn relatively begin collumn section.
     */
    public int cb;
    /**
     * Absolute number of row.
     */
    public int r;
    /**
     * Absolute number of column.
     */
    public int c;
    /**
     * The begining column for BIFF records of types
     * holds general information about columns or range
     * of columns.
     * For (opcode == 125, Column definition, for example)
     */
    public int col_b;
    /**
     * The ending column for BIFF records of types
     * holds general information about columns or range
     * of columns.
     * For (opcode == 125, Column definition, for example)
     */
    public int col_e;
    /**
     * Not used
     */
    public int c1;
    /**
     * Not used
     */
    public int c2;
    /**
     * Not used
     */
    public int w;
    /**
     * Contets of cell.
     * May be one of this objects:
     *   <ul>
     *   <li> null;
     *   <li> Double;
     *   <li> String;
     *   <li> Boolean;
     *   <li> Integer;
     *   </ul>
     */
    public Object d;

  }

  /**
   *
   * <p>Title: Class TSection represents VK Reports Section.</p>
   * <p>Description: The object of this class
   * full describe one report section (verticaly or horizontaly).</p>
   * <p>Copyright: Copyright (c) 2003</p>
   * <p>Company: </p>
   * @author not attributable
   * @version 1.0
   */
  protected class TSection {

    /**
     * Constructor.
     * Initialise empty object.
     */
    public TSection() {
      contents = new Vector();
      clear();
    }

    /**
     * Clean up section object for repeat use.
     */
    public void clear(){
      sectionName = null;
      sectionNum = 0;
      sectionStartPosition = 0;
      sectionLength = 0;
      sectionUsed = false;
      contents.clear();
      rowsCount = 0;
      colsCount = 0;
      widths = null;
    }

    /**
     * Name of section.
     * There are 13 default sections with default names:
     * <table border=1 width=100%>
     *   <tr>
     *     <td><b>Section Names for VK Reports</b></td>
     *     <td><b>For txt, HTML</b></td>
     *     <td><b>For RTF</b></td>
     *     <td><b>For XLS4 vertical</b></td>
     *     <td><b>For XLS4 horisontal</b></td>
     *   </tr>
     *   <tr>
     *
     *     <td>FILE HEADER</td>
     *     <td>-</td>
     *     <td>-</td>
     *     <td>-</td>
     *     <td>-</td>
     *   </tr>
     *   <tr>
     *
     *     <td>REPORT HEADER</td>
     *     <td>\lquote REPORT HEADER\rquote</td>
     *     <td>'REPORT HEADER'</td>
     *     <td>REPORTHEADER</td>
     *     <td>H_REPORTHEADER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>GROUP HEADER</td>
     *     <td>\lquote GROUP HEADER\rquote</td>
     *     <td>'GROUP HEADER'</td>
     *     <td>GROUPHEADER</td>
     *     <td>H_GROUP HEADER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>SUBGROUP HEADER</td>
     *     <td>\lquote SUBGROUP HEADER\rquote</td>
     *     <td>'SUBGROUP HEADER'</td>
     *     <td>SUBGROUPHEADER</td>
     *     <td>H_SUBGROUPHEADER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>SUBSUBGROUP HEADER</td>
     *     <td>\lquote SUBSUBGROUP HEADER\rquote</td>
     *     <td>'SUBSUBGROUP HEADER'</td>
     *     <td>SUBSUBGROUPHEADER</td>
     *     <td>H_SUBSUBGROUPHEADER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>SUPERSUBGROUP HEADER</td>
     *     <td>\lquote SUPERSUBGROUP HEADER\rquote</td>
     *     <td>'SUPERSUBGROUP HEADER'</td>
     *     <td>SUPERSUBGROUPHEADER</td>
     *     <td>H_SUPERSUBGROUPHEADER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>BODY</td>
     *     <td>\lquote BODY\rquote</td>
     *     <td>'BODY'</td>
     *     <td>BODY</td>
     *     <td>H_BODY</td>
     *   </tr>
     *   <tr>
     *
     *     <td>SUPERSUBGROUP FOOTER</td>
     *     <td>\lquote SUPERSUBGROUP FOOTER\rquote</td>
     *     <td>'SUPERSUBGROUP FOOTER'</td>
     *     <td>SUPERSUBGROUPFOOTER</td>
     *     <td>H_SUPERSUBGROUPFOOTER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>SUBSUBGROUP FOOTER</td>
     *     <td>\lquote SUBSUBGROUP FOOTER\rquote</td>
     *     <td>'SUBSUBGROUP FOOTER'</td>
     *     <td>SUBSUBGROUPFOOTER</td>
     *     <td>H_SUBSUBGROUPFOOTER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>SUBGROUP FOOTER</td>
     *     <td>\lquote SUBGROUP FOOTER\rquote</td>
     *     <td>'SUBGROUPFOOTER'</td>
     *     <td>SUBGROUPFOOTER</td>
     *     <td>H_SUBGROUPFOOTER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>GROUP FOOTER</td>
     *     <td>\lquote GROUP FOOTER\rquote</td>
     *     <td>'GROUP FOOTER'</td>
     *     <td>GROUPFOOTER</td>
     *     <td>H_GROUPFOOTER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>REPORT FOOTER</td>
     *     <td>\lquote REPORT FOOTER\rquote</td>
     *     <td>'REPORT FOOTER'</td>
     *     <td>REPORTFOOTER</td>
     *     <td>H_REPORTFOOTER</td>
     *   </tr>
     *   <tr>
     *
     *     <td>FILE FOOTER</td>
     *     <td>-</td>
     *     <td>-</td>
     *     <td>-</td>
     *     <td>-</td>
     *
     *   </tr>
     * </table>
     *
     * You can define your own section with any name in
     * doSectionPrepare event.
     */
    public String sectionName;
    /**
     * Number of section.
     * 0 - 12 default section number, and your own defined
     * since 13.
     * Each section number have static final constant with
     * name:
     *   <ul>
     *   <li> FILE_HEADER  	        = 0;
     *   <li> REPORT_HEADER 	        = 1;
     *   <li> GROUP_HEADER		= 2;
     *   <li> SUBGROUP_HEADER	        = 3;
     *   <li> SUBSUBGROUP_HEADER	= 4;
     *   <li> SUPERSUBGROUP_HEADER	= 5;
     *   <li> BODY			= 6;
     *   <li> SUPERSUBGROUP_FOOTER	= 7;
     *   <li> SUBSUBGROUP_FOOTER	= 8;
     *   <li> SUBGROUP_FOOTER	        = 9;
     *   <li> GROUP_FOOTER		= 10;
     *   <li> REPORT_FOOTER	        = 11;
     *   <li> FILE_FOOTER  	        = 12;
     *   </ul>
     */
    public int sectionNum;
    /**
     * For txt, HTML, RTF report this variable define offset
     * in blank file of start section.
     */
    public int sectionStartPosition;
    /**
     * For txt, HTML, RTF report this variable define length
     * section in bytes.
     */
    public int sectionLength;
    /**
     * Before make report VK Reports generator parse blank
     * and define for each default section used it in report or not.
     * Varable sectionUsed show this one.
     */
    public boolean sectionUsed;
    /**
     * The array of BIFF (TBiffRec objects) records which
     * consists of section.
     * Used only for XLS4 report.
     */
    public Vector contents;
    /**
     * How many rows keep section.
     * Used only for XLS4 report.
     */
    public int rowsCount;
    /**
     * How many columns keep section.
     * Used only for XLS4 report.
     */
    public int colsCount;
    /**
     * Not used
     */
    public int[] widths;

  };

  /**
   *
   * <p>Title: Class BOutStream byte array output stream</p>
   * <p>Description: Used for keep intermediate result
   * of horisontal report.</p>
   * <p>Copyright: Copyright (c) 2003</p>
   * <p>Company: </p>
   * @author not attributable
   * @version 1.0
   */
  protected class BOutStream extends ByteArrayOutputStream {

    /**
     * Constructor. Create and initialise object.
     * @param size
     */
    public BOutStream(int size){
      super(size);
    }

    /**
     * For further use after save an intermediate result
     * report need access to the internal buffer.
     * @return
     */
    public byte[] getBuffer(){
      return buf;
    }

  }

  /**
   * Number of 'FILE HEADER' section.
   */
  public static final int FILE_HEADER  	        = 0;
  /**
   * Number of 'REPORT HEADER' section.
   */
  public static final int REPORT_HEADER 	= 1;
  /**
   * Number of 'GROUP HEADER' section.
   */
  public static final int GROUP_HEADER		= 2;
  /**
   * Number of 'SUBGROUP HEADER' section.
   */
  public static final int SUBGROUP_HEADER	= 3;
  /**
   * Number of 'SUBSUBGROUP HEADER' section.
   */
  public static final int SUBSUBGROUP_HEADER	= 4;
  /**
   * Number of 'SUPERSUBGROUP HEADER' section.
   */
  public static final int SUPERSUBGROUP_HEADER	= 5;
  /**
   * Number of 'BODY' section.
   */
  public static final int BODY			= 6;
  /**
   * Number of 'SUPERSUBGROUP FOOTER' section.
   */
  public static final int SUPERSUBGROUP_FOOTER	= 7;
  /**
   * Number of 'SUBSUBGROUP FOOTER' section.
   */
  public static final int SUBSUBGROUP_FOOTER	= 8;
  /**
   * Number of 'SUBGROUP FOOTER' section.
   */
  public static final int SUBGROUP_FOOTER	= 9;
  /**
   * Number of 'GROUP FOOTER' section.
   */
  public static final int GROUP_FOOTER		= 10;
  /**
   * Number of 'REPORT FOOTER' section.
   */
  public static final int REPORT_FOOTER	        = 11;
  /**
   * Number of 'FILE FOOTER' section.
   */
  public static final int FILE_FOOTER  	        = 12;
  /**
   * The number which is begining of your own section definition.
   */
  public static final int LAST_DYN_SECTION      = 13;
  /**
   * Maximum number of sections.
   */
  public static final int MAX_SECTION_NUM       = 128;
  /**
   * Default symbol for data request by number.
   * In an input blank (stream) you can define mark
   * request data by number like that:
   * &0001, &0002... or @0001, @0002.
   * The symbol '&' or '@' is symbol request data by number.
   * The number after this symbol must be 4 digits.
   * This number pass to the RequestByNum event in
   * VKRepEvent object and you should return with this object
   * the real data for report.
   * Symbol '&' use in XLS4 reports and '@' in TXT, HTML, RTF
   * reports.
   */
  public static final int REQUEST_DATA_BY_NUM_SYMBOL  = 0x26; //&
  /**
   * Default symbol for data request by name.
   * In an input blank (stream) you can define mark
   * request data by name like that:
   * ~Name1~, ~Name2~...
   * The symbol '~' is symbol request data by name.
   * This Data Name closed this symbol pass to the RequestByName
   * event in VKRepEvent object and you should return with this
   * object the real data for report.
   */
  public static final int REQUEST_DATA_BY_NAME_SYMBOL = 0x7E; //~
  /**
   * Used for RTF report. Define begin RTF section breaked.
   */
  public static final int RTF_SECTION_BEGIN_BREAKED   = 0x7B;  // {
  /**
   * Used for RTF report. Define end RTF section breaked.
   */
  public static final int RTF_SECTION_END_BREAKED     = 0x7D;  // }
  /**
   * Maximum length of Data Name for request by name.
   */
  public static final int MAX_LABEL_NAME_LEGTH        =   80;
  /**
   * VK Reports generator load input blank to the memory.
   * STREAM_BUFFER_SIZE is the maximun size of this memory.
   * It meens that you input blank (stream) must be less then
   * STREAM_BUFFER_SIZE constant.
   */
  public static final int STREAM_BUFFER_SIZE          =   256 * 1024;

  /**
   * Array with objects which implemented VKRepListener interface
   * and connected to TReportAbstract object.
   * This field is standard part of java event system.
   * Work if listener field is not null.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   */
  private transient Vector VKRepListeners;
  /**
   * Object which implements VKRepListener interface.
   * It the same VKRepListeners array field, but it is not an
   * array and if listener != null TReportAbstract prefer
   * listener field instead of VKRepListeners field.
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   */
  protected VKRepListener listener;

  /**
   * Array of TSection object.
   * The maximum objects in this array defines MAX_SECTION_NUM constant.
   * @see <a href="TReportAbstract.html#MAX_SECTION_NUM">TReportAbstract.MAX_SECTION_NUM</a>
   */
  protected TSection[] sections;
  /**
   * This variable hold last number dynamic user define section.
   */
  protected int lastDynSection;
  /**
   * Counter for output buffer.
   */
  protected int outputBufferCount;
  /**
   * Variable charByName hold symbol for data request by name.
   * @see <a href="TReportAbstract.html#REQUEST_DATA_BY_NAME_SYMBOL">TReportAbstract.REQUEST_DATA_BY_NAME_SYMBOL</a>
   */
  protected byte charByName = 126; //~
  /**
   * Variable charByName hold symbol for data request by num.
   * @see <a href="TReportAbstract.html#REQUEST_DATA_BY_NUM_SYMBOL">TReportAbstract.REQUEST_DATA_BY_NUM_SYMBOL</a>
   */
  protected byte charByNum = 64;   //@
  /**
   * File name for input blank.
   * Used if blankInputStream is null.
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   */
  protected String inputBlank;
  /**
   * File name for output report.
   * Used if reportOutStream is null.
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   */
  protected String outputReport;
  /**
   * Size of inputBuffer and outputBuffer.
   * When invoke an execute() method VK Reports allocate
   * two buffers. One for input (inputBuffer) and other is
   * for output (outputBuffer).
   * The size of both is streamBufferSize by default is
   * STREAM_BUFFER_SIZE.
   * Input buffer fill full input blank from inputBlank file or
   * blankInputStream stream (prefer blankInputStream if not is null).
   * Output buffer fill gradually (with incremented outputBufferCount
   * variable) and flush to the output file
   * (outputReport) or output stream (reportOutStream) (prefer
   * reportOutStream if not is null) when the output buffer is full.
   * @see <a href="TReportAbstract.html#inputBuffer">TReportAbstract.inputBuffer</a>
   * @see <a href="TReportAbstract.html#outputBuffer">TReportAbstract.outputBuffer</a>
   * @see <a href="TReportAbstract.html#STREAM_BUFFER_SIZE">TReportAbstract.STREAM_BUFFER_SIZE</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.inputBlank</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.outputReport</a>
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   * @see <a href="TReportAbstract.html#outputBufferCount">TReportAbstract.outputBufferCount</a>
   * @see <a href="TReportAbstract.html#execute()">TReportAbstract.execute()</a>
   */
  protected int streamBufferSize = STREAM_BUFFER_SIZE;
  /**
   * Input buffer.
   * Input blank (file or stream) must be less then STREAM_BUFFER_SIZE.
   * @see <a href="TReportAbstract.html#STREAM_BUFFER_SIZE">TReportAbstract.STREAM_BUFFER_SIZE</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.inputBlank</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.outputReport</a>
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   * @see <a href="TReportAbstract.html#outputBufferCount">TReportAbstract.outputBufferCount</a>
   * @see <a href="TReportAbstract.html#streamBufferSize">TReportAbstract.streamBufferSize</a>
   */
  protected byte[] inputBuffer;
  /**
   * Output buffer.
   * Output buffer fill gradually (with incremented outputBufferCount
   * variable) and flush to the output file
   * (outputReport) or output stream (reportOutStream) (prefer
   * reportOutStream if not is null) when the output buffer is full.
   * @see <a href="TReportAbstract.html#STREAM_BUFFER_SIZE">TReportAbstract.STREAM_BUFFER_SIZE</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.inputBlank</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.outputReport</a>
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   * @see <a href="TReportAbstract.html#outputBufferCount">TReportAbstract.outputBufferCount</a>
   * @see <a href="TReportAbstract.html#streamBufferSize">TReportAbstract.streamBufferSize</a>
   */
  protected byte[] outputBuffer;
  /**
   * Real input blank (file or stream) length.
   * Fill after the input blank load to the
   * input buffer inputBuffer.
   * realInputBlankLength must be less then
   * STREAM_BUFFER_SIZE constant.
   * @see <a href="TReportAbstract.html#STREAM_BUFFER_SIZE">TReportAbstract.STREAM_BUFFER_SIZE</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.inputBlank</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.outputReport</a>
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   * @see <a href="TReportAbstract.html#outputBufferCount">TReportAbstract.outputBufferCount</a>
   * @see <a href="TReportAbstract.html#streamBufferSize">TReportAbstract.streamBufferSize</a>
   */
  protected int realInputBlankLength;
  /**
   * Blank input stream. If defined then prefer
   * blankInputStream instead of inputBlank file name
   * for source blank.
   * @see <a href="TReportAbstract.html#inputBuffer">TReportAbstract.inputBuffer</a>
   * @see <a href="TReportAbstract.html#outputBuffer">TReportAbstract.outputBuffer</a>
   * @see <a href="TReportAbstract.html#STREAM_BUFFER_SIZE">TReportAbstract.STREAM_BUFFER_SIZE</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.inputBlank</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.outputReport</a>
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   * @see <a href="TReportAbstract.html#outputBufferCount">TReportAbstract.outputBufferCount</a>
   */
  protected java.io.InputStream blankInputStream;
  /**
   * Report output stream. If defined then prefer
   * reportOutStream instead of outputReport file name
   * for result report.
   * @see <a href="TReportAbstract.html#inputBuffer">TReportAbstract.inputBuffer</a>
   * @see <a href="TReportAbstract.html#outputBuffer">TReportAbstract.outputBuffer</a>
   * @see <a href="TReportAbstract.html#STREAM_BUFFER_SIZE">TReportAbstract.STREAM_BUFFER_SIZE</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.inputBlank</a>
   * @see <a href="TReportAbstract.html#">TReportAbstract.outputReport</a>
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   * @see <a href="TReportAbstract.html#outputBufferCount">TReportAbstract.outputBufferCount</a>
   */
  protected java.io.OutputStream reportOutStream;
  /**
   * This is Event object that pass to all Event handlers.
   */
  protected VKRepEvent rptEvent;
  /**
   * It is work variable input XLS4 file counter.
   * For read information from XLS4 input stream.
   */
  protected int offsetXLSInput = 0;
  /**
   * Horisontal XLS4 report consist of two parts.
   * This pats merge in output stream in end work.
   * BOutStreamA is first of this part.
   * @see <a href="TReportAbstract.html#BOutStreamB">TReportAbstract.BOutStreamB</a>
   */
  protected BOutStream BOutStreamA;
  /**
   * Horisontal XLS4 report consist of two parts.
   * This pats merge in output stream in end work.
   * BOutStreamB is second of this part.
   * @see <a href="TReportAbstract.html#BOutStreamA">TReportAbstract.BOutStreamA</a>
   */
  protected BOutStream BOutStreamB;
  /**
   * Horisontal XLS4 report consist of two parts.
   * First part BOutStreamA, the second BOutStreamB.
   * The variable outCase define in which stream
   * in that moment will be out of output report.
   */
  protected int outCase = 0;

  /**
   * Dimantion XLS4 BIFF record.
   */
  protected short[] XLSDimension = {0x200, 0x0A, 0x0, (short)0xFFFF, 0x0, 0xFF, 0};
  /**
   * Column XLS4 BIFF record.
   */
  protected short[] XLSCOL = {0x7D, 0x0C, 0x0, 0x0, 0x0, 0xF, 0x0, 0x0};
  /**
   * Row XLS4 BIFF record.
   */
  protected short[] XLSRow = {0x208, 16, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0};
  /**
   * Null (or blank) XLS4 BIFF record.
   */
  protected short[] XLSBlank = {0x201, 6, 0x0, 0x0, 0x0};
  /**
   * RK XLS4 BIFF record.
   * This is intricate BIFF record with integer or float number.
   */
  protected short[] XLSRK = {0x27E, 10, 0x0, 0x0, 0x0};
  /**
   * Float number XLS4 BIFF record.
   */
  protected short[] XLSNumber = {0x203, 14, 0x0, 0x0, 0x0};
  /**
   * Label XLS4 BIFF record.
   */
  protected short[] XLSLabel = {0x204, 0, 0x0, 0x0, 0x0, 0x0};
  /**
   * Boolean XLS4 BIFF record.
   */
  protected short[] XLSBool = {0x206, 8, 0x0, 0x0, 0x0};

  /**
   * Constructor.
   * Initialise object.
   */
  public TReportAbstract() {
    int i;
    this.listener = null;
    this.blankInputStream = null;
    this.reportOutStream = null;
    this.rptEvent = new VKRepEvent(this);
    sections = new TSection[MAX_SECTION_NUM];
    outCase = 0;
    for (i=0; i<MAX_SECTION_NUM; i++){
      sections[i] = new TSection();
    }
  }

  /**
   * Implements Serializable interface.
   * @param oos
   * @throws IOException
   */
  private void writeObject(ObjectOutputStream oos) throws IOException {
    oos.defaultWriteObject();
  }
  /**
   * Implements Serializable interface.
   * @param ois
   * @throws ClassNotFoundException
   * @throws IOException
   */
  private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
    ois.defaultReadObject();
  }

  /**
   * Add listener in standard Java event system.
   * @param l object which inmplements VKRepListener interface.
   * @see <a href="VKRepListener.html#VKRepListener">VKRepListener interface</a>
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   */
  public synchronized void addVKRepListener(VKRepListener l) {
    Vector v = VKRepListeners == null ? new Vector(2) : (Vector) VKRepListeners.clone();
    if (!v.contains(l)) {
      v.addElement(l);
      VKRepListeners = v;
    }
  }

  /**
   * Remove listener in standard Java event system.
   * @param l object which inmplements VKRepListener interface.
   * @see <a href="VKRepListener.html#VKRepListener">VKRepListener interface</a>
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   */
  public synchronized void removeVKRepListener(VKRepListener l) {
    if (VKRepListeners != null && VKRepListeners.contains(l)) {
      Vector v = (Vector) VKRepListeners.clone();
      v.removeElement(l);
      VKRepListeners = v;
    }
  }

  /**
   * Property method by read of listener object.
   * @return listener object field.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   */
  public synchronized VKRepListener getListener() {
    return listener;
  }

  /**
   * Property method by write of listener object.
   * @param lster The new value of listener object field.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   */
  public synchronized void setListener(VKRepListener lster) {
    listener = lster;
  }

  /**
   * Start report event.
   * This event fire when report is start.
   * When listener object is not null invoke his startReport mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * nothing
   * <!--
   * <ul>
   * <li> sectionName
   * <li> sectionNum
   * </ul>
   * -->
   * There is no to set in income VKRepEvent object.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   */
  protected void fireStartReport(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).startReport(e);
        }
      }
    } else {
      listener.startReport(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireStartReport(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireStartReport(VKRep.VKRepEvent)">TReportAbstract.fireStartReport(VKRepEvent e)</a>
   */
  public void doStartReport(VKRepEvent e) {
    fireStartReport(e);
  }

  /**
   * End report event.
   * This event fire when report is end.
   * When listener object is not null invoke his endReport mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * nothing
   * <!--
   * <ul>
   * <li> sectionName
   * <li> sectionNum
   * </ul>
   * -->
   * There is no to set in income VKRepEvent object.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   */
  protected void fireEndReport(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).endReport(e);
        }
      }
    } else {
      listener.endReport(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireEndReport(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireEndReport(VKRep.VKRepEvent)">TReportAbstract.fireEndReport(VKRepEvent e)</a>
   */
  public void doEndReport(VKRepEvent e){
    fireEndReport(e);
  }

  /**
   * Report error event.
   * This event fire when error occured during report build process.
   * When listener object is not null invoke his reportError mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> errorMessage
   * </ul>
   * There is no to set in income VKRepEvent object.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   */
  protected void fireReportError(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).reportError(e);
        }
      }
    } else {
      listener.reportError(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireReportError(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireReportError(VKRep.VKRepEvent)">TReportAbstract.fireReportError(VKRepEvent e)</a>
   */
  public void doReportError(VKRepEvent e) {
    fireReportError(e);
  }

  /**
   * Move top data source for report event.
   * This event fire when report build process try to move
   * data source for report to the top.
   * When listener object is not null invoke his fireMoveTop mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> recno - content 1;
   * <li> eof - content true;
   * </ul>
   * You should set in income VKRepEvent object eof variable to
   * determinate achieve end of data source for report.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   */
  protected void fireMoveTop(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).moveTop(e);
        }
      }
    } else {
      listener.moveTop(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireMoveTop(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireMoveTop(VKRep.VKRepEvent)">TReportAbstract.fireMoveTop(VKRepEvent e)</a>
   */
  public void doMoveTop(VKRepEvent e){
    fireMoveTop(e);
  }

  /**
   * Move next row of data source for report event.
   * This event fire when report build process try to move current
   * row of data source for report to the next row.
   * When listener object is not null invoke his fireMoveNext mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> recno - content current recno;
   * <li> eof - content last settings;
   * </ul>
   * You should set in income VKRepEvent object eof variable to
   * determinate achieve end of data source for report.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   */
  protected void fireMoveNext(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).moveNext(e);
        }
      }
    } else {
      listener.moveNext(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireMoveNext(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireMoveNext(VKRep.VKRepEvent)">TReportAbstract.fireMoveNext(VKRepEvent e)</a>
   */
  public void doMoveNext(VKRepEvent e){
    fireMoveNext(e);
  }

  /**
   * Section begin report event.
   * This event fire when report build process try to copy report
   * section to the output stream.
   * When listener object is not null invoke his fireSectionBegin mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * </ul>
   * This event only notify main proccess that report process
   * try to copy report section to the output stream and you should
   * not to set any field in VKRepEvent object.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   * @see <a href="TReportAbstract.html#fireSectionBegin(VKRep.VKRepEvent)">TReportAbstract.fireSectionBegin(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionCheck(VKRep.VKRepEvent)">TReportAbstract.fireSectionCheck(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionComplete(VKRep.VKRepEvent)">TReportAbstract.fireSectionComplete(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionEnd(VKRep.VKRepEvent)">TReportAbstract.fireSectionEnd(VKRepEvent e)</a>
   */
  protected void fireSectionBegin(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).sectionBegin(e);
        }
      }
    } else {
      listener.sectionBegin(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireSectionBegin(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireSectionBegin(VKRep.VKRepEvent)">TReportAbstract.fireSectionBegin(VKRepEvent e)</a>
   */
  public void doSectionBegin(VKRepEvent e){
    fireSectionBegin(e);
  }

  /**
   * Section end report event.
   * This event fire when report build process end to copy report
   * section to the output stream.
   * This event trigger always, independently it was copy report
   * section to the output stream or not.
   * When listener object is not null invoke his fireSectionEnd mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * </ul>
   * This event only notify main proccess that report process
   * end to copy report section to the output stream and you should
   * not to set any field in VKRepEvent object.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   * @see <a href="TReportAbstract.html#fireSectionBegin(VKRep.VKRepEvent)">TReportAbstract.fireSectionBegin(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionCheck(VKRep.VKRepEvent)">TReportAbstract.fireSectionCheck(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionComplete(VKRep.VKRepEvent)">TReportAbstract.fireSectionComplete(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionEnd(VKRep.VKRepEvent)">TReportAbstract.fireSectionEnd(VKRepEvent e)</a>
   */
  protected void fireSectionEnd(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).sectionEnd(e);
        }
      }
    } else {
      listener.sectionEnd(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireSectionEnd(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireSectionEnd(VKRep.VKRepEvent)">TReportAbstract.fireSectionEnd(VKRepEvent e)</a>
   */
  public void doSectionEnd(VKRepEvent e){
    fireSectionEnd(e);
  }

  /**
   * Section complete report event.
   * This event fire when report build process complete to copy report
   * section to the output stream.
   * This event trigger only if report section was realy copy
   * to the output stream.
   * When listener object is not null invoke his fireSectionComplete mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * </ul>
   * This event only notify main proccess that report process
   * complete to copy report section to the output stream and you should
   * not to set any field in VKRepEvent object.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   * @see <a href="TReportAbstract.html#fireSectionBegin(VKRep.VKRepEvent)">TReportAbstract.fireSectionBegin(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionCheck(VKRep.VKRepEvent)">TReportAbstract.fireSectionCheck(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionComplete(VKRep.VKRepEvent)">TReportAbstract.fireSectionComplete(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionEnd(VKRep.VKRepEvent)">TReportAbstract.fireSectionEnd(VKRepEvent e)</a>
   */
  protected void fireSectionComplete(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).sectionComplete(e);
        }
      }
    } else {
      listener.sectionComplete(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireSectionComplete(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireSectionComplete(VKRep.VKRepEvent)">TReportAbstract.fireSectionComplete(VKRepEvent e)</a>
   */
  public void doSectionComplete(VKRepEvent e){
    fireSectionComplete(e);
  }

  /**
   * Section check report event.
   * This event fire when report build process try to copy report
   * section to the output stream and wants to know it will be
   * realy need to copy or not.
   * When listener object is not null invoke his fireSectionCheck mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.
   * In income VKRepEvent object impotant field:
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * <li> check - the flag of copy report section. It default value
   * depends from section:
   *   <ul>
   *     <li> FILE_HEADER  	        true;
   *     <li> REPORT_HEADER 	        true;
   *     <li> GROUP_HEADER		false;
   *     <li> SUBGROUP_HEADER	        false;
   *     <li> SUBSUBGROUP_HEADER	false;
   *     <li> SUPERSUBGROUP_HEADER	false;
   *     <li> BODY			true;
   *     <li> SUPERSUBGROUP_FOOTER	false;
   *     <li> SUBSUBGROUP_FOOTER	false;
   *     <li> SUBGROUP_FOOTER	        false;
   *     <li> GROUP_FOOTER		false;
   *     <li> REPORT_FOOTER	        true;
   *     <li> FILE_FOOTER  	        true;
   *   </ul>
   * </ul>
   * When you catch this event you should to set VKRepEvent.check
   * field to true if you want to copy report section in present time
   * or false if not.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#listener">TReportAbstract.listener</a>
   * @see <a href="TReportAbstract.html#VKRepListeners">TReportAbstract.VKRepListeners</a>
   * @see <a href="TReportAbstract.html#fireSectionBegin(VKRep.VKRepEvent)">TReportAbstract.fireSectionBegin(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionCheck(VKRep.VKRepEvent)">TReportAbstract.fireSectionCheck(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionComplete(VKRep.VKRepEvent)">TReportAbstract.fireSectionComplete(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireSectionEnd(VKRep.VKRepEvent)">TReportAbstract.fireSectionEnd(VKRepEvent e)</a>
   */
  protected void fireSectionCheck(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).sectionCheck(e);
        }
      }
    } else {
      listener.sectionCheck(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireSectionCheck(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireSectionCheck(VKRep.VKRepEvent)">TReportAbstract.fireSectionCheck(VKRepEvent e)</a>
   */
  public void doSectionCheck(VKRepEvent e){
    fireSectionCheck(e);
  }

  /**
   * <p>Data request report event.</p>
   * <p>This event fire when report build process need a data for report.
   * Generaly data request event used in grid based reports like a
   * XLS4 report (TXLS4ReportVertical, TXLS4ReportHorizontal).
   * For this type reports data request event query data
   * for a cells of grid. For a text based reports (txt, html, rtf
   * - TWordReport) this event not used.</p>
   * <p>When listener object is not null invoke his fireDataRequest mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.</p>
   * <p>In income VKRepEvent object impotant field:</p>
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * <li> check - the flag data changed;
   * <li> rowAbs - absolute row of cell in output report;
   * <li> colAbs - absolute column of cell in output report;
   * <li> rowRel - row of cell relatively the begin of report section;
   * <li> colRel - column of cell relatively the begin of report section;
   * <li> dataInput - the object represent the content of cell. May be
   * one of following types:
   *   <ul>
   *     <li> null;
   *     <li> Double;
   *     <li> String;
   *     <li> Boolean;
   *     <li> Integer;
   *   </ul>
   * The Date or Timestamp in XLS4 format represent as Double.
   * The integer part is amount of days since 01/01/1900 and
   * the fractional part is part of day (time);
   * <li> dataOutput - the object represent the new content of cell.
   * Types of object the same as for dataInput;
   * <li> outData - the byte array for output;
   * <li> outDataLen - amount of bytes in outData field for output;
   * </ul>
   * In respond of this event you should fill data for out and
   * set check to true (check = true let knows VK Reports that
   * you want to fill cell with a new value) and fill dataOutput field
   * or outData/outDataLen pair.
   * If you want to clear output cell (assign to null) you should set
   * dataOutput = null and outDataLen = 0.
   * The String and any binary data much better to pass into output
   * report with outData/outDataLen pair.
   * In Excel 4 maximum length of cell is 255 bytes.
   * The maximum length of outData array define in VKRepEvent class.
   * If dataOutput != null and outDataLen != 0 then VK Reports prefer
   * dataOutput field.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireRequestByName(VKRep.VKRepEvent)">TReportAbstract.fireRequestByName(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireRequestByNum(VKRep.VKRepEvent)">TReportAbstract.fireRequestByNum(VKRepEvent e)</a>
   */
  protected void fireDataRequest(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).dataRequest(e);
        }
      }
    } else {
      listener.dataRequest(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireDataRequest(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireDataRequest(VKRep.VKRepEvent)">TReportAbstract.fireDataRequest(VKRepEvent e)</a>
   */
  public void doDataRequest(VKRepEvent e){
    fireDataRequest(e);
  }

  /**
   * <p>Data request by name report event.</p>
   * <p>This event fire when report build process need a data for report.
   * Data request event by name used in grid based reports like a
   * XLS4 report (TXLS4ReportVertical, TXLS4ReportHorizontal).
   * For this type reports data request event query data
   * for a part of cells of grid.</p>
   * <p>And used for a text based reports (txt, html, rtf - TWordReport).</p>
   * <p>When listener object is not null invoke his fireDataRequest mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.</p>
   * <p>In income VKRepEvent object impotant field:</p>
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * <li> dataName - name of data. In input blank(stream) the
   * DataName fields define like that: ~DataName1~, ~DataName2~;
   * Symbol '~' define in property of specific report component.
   * <li> outData - the byte array for output;
   * <li> outDataLen - amount of bytes in outData field for output;
   * </ul>
   * In respond of this event you should fill outData/outDataLen
   * pair.
   * If you want to clear output data (assign to null) you should set
   * outDataLen = 0.
   * The maximum length of outData array define in VKRepEvent class.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireDataRequest(VKRep.VKRepEvent)">TReportAbstract.fireDataRequest(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireRequestByNum(VKRep.VKRepEvent)">TReportAbstract.fireRequestByNum(VKRepEvent e)</a>
   */
  protected void fireRequestByName(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).requestByName(e);
        }
      }
    } else {
      listener.requestByName(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireRequestByName(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireRequestByName(VKRep.VKRepEvent)">TReportAbstract.fireRequestByName(VKRepEvent e)</a>
   */
  public void doRequestByName(VKRepEvent e){
    fireRequestByName(e);
  }

  /**
   * <p>Data request by number report event.</p>
   * <p>This event fire when report build process need a data for report.
   * Data request event by number used in grid based reports like a
   * XLS4 report (TXLS4ReportVertical, TXLS4ReportHorizontal).
   * For this type reports data request event query data
   * for a part of cells of grid.</p>
   * <p>And used for a text based reports (txt, html, rtf - TWordReport).</p>
   * <p>When listener object is not null invoke his fireDataRequest mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.</p>
   * <p>In income VKRepEvent object impotant field:</p>
   * <ul>
   * <li> sectionName - name of report section;
   * <li> sectionNum - number of report section;
   * <li> dataIndex - index (number) of data. In input blank(stream) the
   * DataNumber fields define like that: @0001, @0002, ...
   * or &0001, &0002;
   * Symbols '@' or '&' define in property of specific report component.
   * <li> outData - the byte array for output;
   * <li> outDataLen - amount of bytes in outData field for output;
   * </ul>
   * In respond of this event you should fill outData/outDataLen
   * pair.
   * If you want to clear output data (assign to null) you should set
   * outDataLen = 0.
   * The maximum length of outData array define in VKRepEvent class.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireDataRequest(VKRep.VKRepEvent)">TReportAbstract.fireDataRequest(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#fireRequestByName(VKRep.VKRepEvent)">TReportAbstract.fireRequestByName(VKRepEvent e)</a>
   */
  protected void fireRequestByNum(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).requestByNum(e);
        }
      }
    } else {
      listener.requestByNum(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireRequestByNum(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireRequestByNum(VKRep.VKRepEvent)">TReportAbstract.fireRequestByNum(VKRepEvent e)</a>
   */
  public void doRequestByNum(VKRepEvent e) {
    fireRequestByNum(e);
  }

  /**
   * <p>Section prepare report event.</p>
   * <p>This event fire when report build process has made parse
   * input blank(stream) and need to know would you like to define
   * your own sections extra of 13 standards.
   * You can define them to invoke prepareSection(String SectionName)
   * method with supply it your own section name which you
   * describe in input blank (stream).</p>
   * <p>When listener object is not null invoke his fireSectionPrepare mothod.
   * If not, switch on standard Java fire event throw VKRepListeners variable.</p>
   * <p>In income VKRepEvent object there is no impotant field for this event.</p>
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#prepareSection(java.lang.String)">TReportAbstract.prepareSection(String SectionName)</a>
   */
  protected void fireSectionPrepare(VKRepEvent e) {
    if ( listener == null ) {
      if (VKRepListeners != null) {
        Vector listeners = VKRepListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ( (VKRepListener) listeners.elementAt(i)).sectionPrepare(e);
        }
      }
    } else {
      listener.sectionPrepare(e);
    }
  }

  /**
   * This public method only invoke TReportAbstract.fireSectionPrepare(VKRepEvent e).
   * You can override this method in descendant classes and handle
   * this event in you method.
   * @param e VKRepEvent event object.
   * @see <a href="TReportAbstract.html#fireSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.fireSectionPrepare(VKRepEvent e)</a>
   */
  public void doSectionPrepare(VKRepEvent e) {
    fireSectionPrepare(e);
  }

  /**
   * Out report section to the output report (file or stream)
   * by the name of section.
   * You can use this method when you define your own order
   * sections in your report instead of standard order.
   * You can out any section (including your own, prepared
   * in SectionPrepare event or method) in any time.
   * @param SectionName Section Name
   * @see <a href="TReportAbstract.html#fireSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.fireSectionPrepare(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#doSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.doSectionPrepare(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#outSection(int)">TReportAbstract.outSection(int)</a>
   */
  public void outSection(String SectionName){
    int i;
    int j = -1;
    loop: {
      for (i = 0; i < MAX_SECTION_NUM; i++) {
        if (sections[i].sectionName.equals(SectionName)) {
          j = i;
          break loop;
        }
      }
    }
    if ( j > 0 ) {
      if (sections[j].sectionUsed) {
        outSection(j);
      }
    } else {
      this.rptEvent.errorMessage = "There is no section '" + SectionName + "' for out.";
      doReportError(this.rptEvent);
    }
  };

  /**
   * Out report section to the output report (file or stream)
   * by the number of section.
   * You can use this method when you define your own order
   * sections in your report instead of standard order.
   * You can out any section (including your own, prepared
   * in SectionPrepare event or method) in any time.
   * @param SectionNum Number of section
   * @see <a href="TReportAbstract.html#fireSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.fireSectionPrepare(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#doSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.doSectionPrepare(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#outSection(java.lang.String)">TReportAbstract.outSection(java.lang.String)</a>
   */
  public void outSection(int SectionNum){
    outSectionInternal(SectionNum, false);
  };

  /**
   * Method execute start process of build report.
   */
  public void execute(){
    doStartReport(this.rptEvent);
    try{
      inputBuffer = new byte[streamBufferSize];
      outputBuffer = new byte[streamBufferSize];
      try {
        if (openInputStream()) {
          try {
            if (openOutputStream()) {
              try {
                try {
                  realInputBlankLength = this.blankInputStream.read(inputBuffer);
                } catch (Exception ee) {
                  realInputBlankLength = 0;
                  this.rptEvent.errorMessage = "Read input stream error.";
                  doReportError(this.rptEvent);
                }
                clearSections();
                lastDynSection = LAST_DYN_SECTION;
                if (prepareFileHeader()) {
                  prepareSectionInternal("REPORT HEADER", REPORT_HEADER);
                  prepareSectionInternal("GROUP HEADER", GROUP_HEADER);
                  prepareSectionInternal("SUBGROUP HEADER", SUBGROUP_HEADER);
                  prepareSectionInternal("SUBSUBGROUP HEADER", SUBSUBGROUP_HEADER);
                  prepareSectionInternal("SUPERSUBGROUP HEADER", SUPERSUBGROUP_HEADER);
                  prepareSectionInternal("BODY", BODY);
                  prepareSectionInternal("SUPERSUBGROUP FOOTER", SUPERSUBGROUP_FOOTER);
                  prepareSectionInternal("SUBSUBGROUP FOOTER", SUBSUBGROUP_FOOTER);
                  prepareSectionInternal("SUBGROUP FOOTER", SUBGROUP_FOOTER);
                  prepareSectionInternal("GROUP FOOTER", GROUP_FOOTER);
                  prepareSectionInternal("REPORT FOOTER", REPORT_FOOTER);
                  doSectionPrepare(this.rptEvent);
                  if (prepareFileFooter()) {
                    outSectionInternal(FILE_HEADER, true);
                    clearVars();
                    outSectionInternal(REPORT_HEADER, true);
                    this.rptEvent.recno = 1;
                    this.rptEvent.eof = true;
                    doMoveTop(this.rptEvent);
                    while (!this.rptEvent.eof) {
                      outSectionInternal(GROUP_HEADER, false);
                      outSectionInternal(SUBGROUP_HEADER, false);
                      outSectionInternal(SUBSUBGROUP_HEADER, false);
                      outSectionInternal(SUPERSUBGROUP_HEADER, false);
                      outSectionInternal(BODY, true);
                      outSectionInternal(SUPERSUBGROUP_FOOTER, false);
                      outSectionInternal(SUBSUBGROUP_FOOTER, false);
                      outSectionInternal(SUBGROUP_FOOTER, false);
                      outSectionInternal(GROUP_FOOTER, false);
                      doMoveNext(this.rptEvent);
                      this.rptEvent.recno++;
                    }
                    outSectionInternal(REPORT_FOOTER, true);
                    outSectionInternal(FILE_FOOTER, true);
                    outSpecial();
                  }
                }
              } finally {
                closeOutputStream();
              }
            } else {
              this.rptEvent.errorMessage = "Open output stream error.";
              doReportError(this.rptEvent);
            }
          } finally {
            closeInputStream();
          }
        } else {
          this.rptEvent.errorMessage = "Open input stream error.";
          doReportError(this.rptEvent);
        }
      } finally {
        inputBuffer = null;
        outputBuffer = null;
      }
    } finally {
      doEndReport(this.rptEvent);
    }
  };

  /**
   * Implements runable interface, call execute() method
   * in Thread.
   * @see <a href="TReportAbstract.html#execute()">TReportAbstract.execute()</a>
   */
  public void run(){
    execute();
  }

  /**
   * Access method by read of blankInputStream property.
   * @return Value of blankInputStream property.
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   */
  public java.io.InputStream getBlankInputStream() {
    return blankInputStream;
  }

  /**
   * Access method by write of blankInputStream property.
   * @param blankInputStream Input stream of blank report.
   * @see <a href="TReportAbstract.html#blankInputStream">TReportAbstract.blankInputStream</a>
   */
  public void setBlankInputStream(java.io.InputStream blankInputStream) {
    this.blankInputStream = blankInputStream;
  }

  /**
   * Access method by read of reportOutStream property.
   * @return Value of reportOutStream property.
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   */
  public java.io.OutputStream getReportOutStream() {
    return reportOutStream;
  }

  /**
   * Access method by write of reportOutStream property.
   * @param reportOutStream Output stream for result report.
   * @see <a href="TReportAbstract.html#reportOutStream">TReportAbstract.reportOutStream</a>
   */
  public void setReportOutStream(java.io.OutputStream reportOutStream) {
    this.reportOutStream = reportOutStream;
  }

  /**
   * Initialize array of sections.
   */
  protected void clearSections(){
    for (int i=0; i<MAX_SECTION_NUM; i++) sections[i].clear();
  }

  /**
   * Virtual auxiliary method.
   */
  protected void clearVars(){
  };

  /**
   * Access method by read of streamBufferSize property.
   * @return Value of streamBufferSize property.
   * @see <a href="TReportAbstract.html#streamBufferSize">TReportAbstract.streamBufferSize</a>
   */
  public int getStreamBufferSize(){
    return streamBufferSize;
  };
  /**
   * Access method by write of streamBufferSize property.
   * @param newStreamBufferSize New value of streamBufferSize
   * property.
   * @see <a href="TReportAbstract.html#streamBufferSize">TReportAbstract.streamBufferSize</a>
   */
  public void setStreamBufferSize(int newStreamBufferSize){
    streamBufferSize = newStreamBufferSize;
  };

  /**
   * Access method by read of charByName property.
   * @return Value of charByName property.
   * @see <a href="TReportAbstract.html#charByName">TReportAbstract.charByName</a>
   */
  public byte getCharByName() {
    return charByName;
  }
  /**
   * Access method by write of charByName property.
   * @param charByName New value of charByName
   * property.
   * @see <a href="TReportAbstract.html#charByName">TReportAbstract.charByName</a>
   */
  public void setCharByName(byte charByName) {
    this.charByName = charByName;
  }

  /**
   * Access method by read of charByNum property.
   * @return Value of charByNum property.
   * @see <a href="TReportAbstract.html#charByNum">TReportAbstract.charByNum</a>
   */
  public byte getCharByNum() {
    return charByNum;
  }
  /**
   * Access method by write of charByNum property.
   * @param charByNum New value of charByNum
   * property.
   * @see <a href="TReportAbstract.html#charByNum">TReportAbstract.charByNum</a>
   */
  public void setCharByNum(byte charByNum) {
    this.charByNum = charByNum;
  }

  /**
   * Access method by read of inputBlank (blank file name) property.
   * @return Value of inputBlank (blank file name) property.
   * @see <a href="TReportAbstract.html#inputBlank">TReportAbstract.inputBlank</a>
   */
  public String getInputBlank() {
    return inputBlank;
  }
  /**
   * Access method by write of inputBlank (blank file name) property.
   * @param inputBlank New value of inputBlank (blank file name)
   * property.
   * @see <a href="TReportAbstract.html#inputBlank">TReportAbstract.inputBlank</a>
   */
  public void setInputBlank(String inputBlank) {
    this.inputBlank = inputBlank;
  }

  /**
   * Access method by read of outputReport (output report file name)
   * property.
   * @return Value of outputReport (output report file name) property.
   * @see <a href="TReportAbstract.html#outputReport">TReportAbstract.outputReport</a>
   */
  public String getOutputReport() {
    return outputReport;
  }
  /**
   * Access method by write of outputReport (output report file
   * name) property.
   * @param outputReport New value of outputReport (output report
   * file name) property.
   * @see <a href="TReportAbstract.html#outputReport">TReportAbstract.outputReport</a>
   */
  public void setOutputReport(String outputReport) {
    this.outputReport = outputReport;
  }

  /**
   * When you define your own section, you describe it
   * in input blank. Then, in doSectionPrepare method or in
   * fireSectionPrepare event you should call prepareSection
   * method for parse section for further output.
   * @param SectionName Your own section name for prepare.
   * @return new section number (greater then last section
   * number) and increment the last section number.
   */
  public int prepareSection(String SectionName){
    return prepareSectionInternal(SectionName, -1);
  };

  /**
   * Write full output buffer to output stream and set
   * outputBufferCount to zerro.
   */
  protected void flushOutBuffer(){
    try {
      this.reportOutStream.write(outputBuffer, 0, outputBufferCount);
      this.reportOutStream.flush();
    }
    catch (Exception ee) {
      this.rptEvent.errorMessage = "Write stream error.";
      doReportError(this.rptEvent);
    }
    outputBufferCount = 0;
  }

  /**
   * Close output buffer. Flow together two auxiliary streams
   * used in XLS4 horisontal report and flush output stream.
   */
  protected void closeOutBuffer(){
    try {
      if ( (BOutStreamA.size() > 0) && (BOutStreamB.size() > 0)) {
        BOutStreamA.writeTo(this.reportOutStream);
        BOutStreamB.writeTo(this.reportOutStream);
      }
      this.reportOutStream.write(outputBuffer, 0, outputBufferCount);
      this.reportOutStream.flush();
    }
    catch (Exception ee) {
      this.rptEvent.errorMessage = "Write stream error.";
      doReportError(this.rptEvent);
    }
    outputBufferCount = 0;
  }

  /**
   * Out in output stream XLS4 Double BIFF record.
   * @param value Double value for output.
   * @param cellFormat reference to the cell format.
   * @param row row of cell.
   * @param col column for cell.
   */
  protected void outXLSDouble(Double value, short cellFormat, int row, int col){
    XLSNumber[2] = (short)(row);   //row cell
    XLSNumber[3] = (short)(col);   //col cell
    XLSNumber[4] = cellFormat;     //format cell
    outShorts(XLSNumber);
    outDouble(value);
  }

  /**
   * Out in output stream BIFF record with any value
   * depends of rptEvent object.
   * @param pp BIFF record from input blank.
   * @param row row of cell.
   * @param col column for cell.
   */
  protected void outVariant(TBiffRec pp, int row, int col){
    short cellFormat = getShort(pp.data, 4);
    int lengthStr;
    int C;
    //NULL
    if ( ( rptEvent.dataOutput == null ) &&
         ( rptEvent.outDataLen == 0 ) ){
      XLSBlank[2] = (short)(row);          //row cell
      XLSBlank[3] = (short)(col);          //col cell
      XLSBlank[4] = cellFormat;            //format cell
      outShorts(XLSBlank);
    }
    if ( rptEvent.dataOutput != null ){
      if ( rptEvent.dataOutput instanceof Integer ){
        C = ((Integer)rptEvent.dataOutput).intValue();
        if ( (-536870912 <= C) && (C <= 536870911) ) {
          C = ( ( C << 2 ) + 2 );
          XLSRK[2] = (short)(row);   //row cell
          XLSRK[3] = (short)(col);   //col cell
          XLSRK[4] = cellFormat;     //format cell
          outShorts(XLSRK);
          outInteger(C);
        } else {
          outXLSDouble(new Double(C), cellFormat, row, col);
        }
      }
      if ( rptEvent.dataOutput instanceof Double ) outXLSDouble((Double)rptEvent.dataOutput, cellFormat, row, col);
      if ( rptEvent.dataOutput instanceof Boolean ){
        XLSBool[2] = (short)(row);
        XLSBool[3] = (short)(col);
        XLSBool[4] = cellFormat;
        outShorts(XLSBool);
        if ( ((Boolean)rptEvent.dataOutput).booleanValue() ) {
          outShort( (short) 1);
        } else {
          outShort( (short) 0);
        }
      }
      if ( rptEvent.dataOutput instanceof String ){
        lengthStr = ((String)rptEvent.dataOutput).length();
        if ( lengthStr > 255 ) lengthStr = 254;
        XLSLabel[1] = (short)(lengthStr + 8);
        XLSLabel[2] = (short)(row);
        XLSLabel[3] = (short)(col);
        XLSLabel[4] = cellFormat;
        XLSLabel[5] = (short)(lengthStr);
        outShorts(XLSLabel);
        outBytes(((String)rptEvent.dataOutput).getBytes(), 0, lengthStr);
      }
    } else {
      // rptEvent.dataOutput == null
      if ( rptEvent.outDataLen != 0 ) {
        lengthStr = rptEvent.outDataLen;
        if ( lengthStr > 254 ) lengthStr = 254;
        XLSLabel[1] = (short)(lengthStr + 8);
        XLSLabel[2] = (short)(row);
        XLSLabel[3] = (short)(col);
        XLSLabel[4] = cellFormat;
        XLSLabel[5] = (short)(lengthStr);
        outShorts(XLSLabel);
        outBytes(rptEvent.outData, 0, lengthStr);
      }
    }
  }

  /**
   * Read bytes from input blank (stream).
   * @param b
   */
  protected void getBytes(byte[] b){
    int i, j = b.length;
    for ( i=0; i<j; i++ ) {
      b[i] = inputBuffer[offsetXLSInput];
      offsetXLSInput++;
    }
  }

  /**
   * Read short value from input blank (stream).
   * @return short value from input blank (stream).
   */
  protected short getShort(){
    short result = (short) ( (((short)inputBuffer[offsetXLSInput + 1] << 8) & (short)0xFF00) +
                             (((short)inputBuffer[offsetXLSInput + 0] << 0) & (short)0x00FF) );
    offsetXLSInput+=2;
    return result;
  }

  /**
   * Read long value from byte array b with position pos.
   * @param b array of bytes for read.
   * @param pos position to read with.
   * @return Long value.
   */
  protected long getLong(byte[] b, int pos){
    return (long) ( (((long)b[pos + 7] << 56) & (long)0xFF00000000000000L) +
                    (((long)b[pos + 6] << 48) & (long)0x00FF000000000000L) +
                    (((long)b[pos + 5] << 40) & (long)0x0000FF0000000000L) +
                    (((long)b[pos + 4] << 32) & (long)0x000000FF00000000L) +
                    (((long)b[pos + 3] << 24) & (long)0x00000000FF000000L) +
                    (((long)b[pos + 2] << 16) & (long)0x0000000000FF0000L) +
                    (((long)b[pos + 1] << 8 ) & (long)0x000000000000FF00L) +
                    (((long)b[pos + 0] << 0 ) & (long)0x00000000000000FFL) );
  }

  /**
   * Read int value from byte array b with position pos.
   * @param b array of bytes for read.
   * @param pos position to read with.
   * @return Int value.
   */
  protected int getInt(byte[] b, int pos){
    return (int) ( (((int)b[pos + 3] << 24) & (int)0xFF000000) +
                   (((int)b[pos + 2] << 16) & (int)0x00FF0000) +
                   (((int)b[pos + 1] << 8 ) & (int)0x0000FF00) +
                   (((int)b[pos + 0] << 0 ) & (int)0x000000FF) );
  }

  /**
   * Read short value from byte array b with position pos.
   * @param b array of bytes for read.
   * @param pos position to read with.
   * @return Short value.
   */
  protected short getShort(byte[] b, int pos){
    return (short) ( (((short)b[pos + 1] << 8) & (short)0xFF00) +
                     (((short)b[pos + 0] << 0) & (short)0x00FF) );
  }

  /**
   * Write short value to byte array b with position pos.
   * @param b array of bytes for write.
   * @param pos position to write with.
   * @param value value to write.
   */
  protected void setShort(byte[] b, int pos, short value){
    b[pos + 1] = (byte)((value >>> 8) & (short)0xFF);
    b[pos + 0] = (byte)((value >>> 0) & (short)0xFF);
  }

  /**
   * Convert byte to short.
   * @param b byte
   * @return short value.
   */
  protected short byteAsShort(byte b){
    return (short) ( ( (short)0x0000) +
                      (b & (short)0x00FF) );
  }

  /**
   * Convert value from XLS4 RC format to number
   * (Integer or Double).
   * @param j 4 bytes XLS4 RC number format.
   * @return Integer or Double object represents of content
   * RC cell.
   */
  public static Object ass2Number(int j){
    Object result;
    boolean sign, q, q1;
    int m, x, p;
    if ( ( j == 0 ) || ( j == 1 ) || ( j == 2 ) || ( j == 3 ) ) {
      result = new Integer(0);
    } else {
      q1 = ((j & 1) == 1);
      if ((j & 2) == 2){
        p = j;
        q = (p == (p | 0x80000000));
        p = (p >>> 2);
        if ( q ) p = (p | 0xC0000000);
        result = new Integer(p);
      } else {
        sign = (j == (j | 0x80000000));
        p = j;
        q = (p == (p | 0x40000000));
        p = ( p & 0x7FFFFFFF );
        p = p >>> 20;
        p = ( p ^ 0x00000400 );
        if ( !q ) {
          try {
            p = (p | 0xfffff800);
          } catch (Exception e) {
            p = -1;
          }
        }
        x = ( p - 17 );
        p = ( j & 0x000FFFFF );
        p = ( p | 0x00100000 );
        m = ( p >>> 2 );
        if ( isInt(j) ) {
          result = new Integer(new Double(m*exp2(x)).intValue());
          if ( sign ) result = new Integer(-((Integer)result).intValue());
        } else {
          result = new Double(m*exp2(x));
          if ( sign ) result = new Double(-((Double)result).doubleValue());
        }
      }
      if ( q1 ){
        if ( result instanceof  Integer) {
          result = new Double(((Integer)result).intValue() * 0.01);
        } else {
          if (result instanceof Double)
            result = new Double( ( (Double) result).doubleValue() * 0.01);
        }
      }
    }
    return result;
  }

  /**
   * Determine is XLS4 RC cell content Integer or Double.
   * @param p XLS4 RC cell content.
   * @return true - Value is integer; false - Value is double.
   */
  public static boolean isInt(int p){
    boolean b, result;
    int q;
    int mpos;
    q = p;
    b = (p == (p | 0x40000000));  //  Second left bit is set
    p = ( p >>> 20 );
    p = ( p ^ 0x00000400 );
    if ( !b ) p = ( p | 0xfffff800 );
    if ( p != 0xFFFFFFFF ) {
      p++;
    } else {
      p = 0;
    }
    q = ( q & 0x000FFFFF );
    q = ( q >>> 1 );
    mpos = 0;
    if ( q > 0 ){
      mpos = 19;
      loop: {
        while (true) {
          mpos--;
          q = (q >>> 1);
          if ( (q & 1) == 1 ) break loop;
        }
      }
    }
    if ( mpos > p ) {
      result = false; //Value is not integer
    } else {
      result = true; //Value is integer
    }
    return result;
  }

  /**
   * 2 power by x.
   * @param x
   * @return
   */
  public static double exp2(int x) {
    boolean sign = false;
    int i;
    double result = 1;
    if ( x < 0 ) {
      sign = true;
      x = -x;
    }
    for (i=0; i<x; i++) result *= 2;
    if ( sign ) result = (1/result);
    return result;
  }

  /**
   * Removs all blanks from the string.
   * @param inStr String with blanks.
   * @return String without blanks.
   */
  public static String removeAllBlanks(String inStr){
    int i, j = inStr.length();
    char[] qq = new char[j];
    int outStrLen = 0;
    char blank = " ".charAt(0);
    char charIn;
    for (i=0; i<j; i++){
      charIn = inStr.charAt(i);
      if ( charIn != blank ) qq[outStrLen++] = charIn;
    }
    return new String(qq, 0, outStrLen);
  }

  /**
   * Out integer value to the output stream.
   * @param intForOut value for output.
   */
  protected void outInteger(int intForOut){
    outByte((byte)((intForOut >>> 0 ) & (int)0x000000FF));
    outByte((byte)((intForOut >>> 8 ) & (int)0x000000FF));
    outByte((byte)((intForOut >>> 16) & (int)0x000000FF));
    outByte((byte)((intForOut >>> 24) & (int)0x000000FF));
  }

  /**
   * Out Double value to the output stream.
   * @param doubleForOut value for output.
   */
  protected void outDouble(Double doubleForOut){
    long longForOut = Double.doubleToLongBits(doubleForOut.doubleValue());
    outByte((byte)((longForOut >>> 0 ) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 8 ) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 16) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 24) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 32) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 40) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 48) & (long)0x00000000000000FFL));
    outByte((byte)((longForOut >>> 56) & (long)0x00000000000000FFL));
  }

  /**
   * Out Short array value to the output stream.
   * @param shortsForOut value for output.
   */
  protected void outShorts(short[] shortsForOut){
    for (int i=0; i<shortsForOut.length; i++){
      outByte((byte)((shortsForOut[i] >>> 0) & (short)0x00FF));
      outByte((byte)((shortsForOut[i] >>> 8) & (short)0x00FF));
    }
  }

  /**
   * Out byte array to the output stream.
   * @param bytesForOut value for output.
   */
  protected void outBytes(byte[] bytesForOut){
    for (int i=0; i<bytesForOut.length; i++) outByte(bytesForOut[i]);
  }

  /**
   * Out piece byte array to the output stream.
   * @param bytesForOut byte array.
   * @param offset offset.
   * @param len length for out.
   */
  protected void outBytes(byte[] bytesForOut, int offset, int len){
    for (int i=0; i<len; i++) outByte(bytesForOut[offset + i]);
  }

  /**
   * Out byte to the output stream.
   * @param byteForOut value for output.
   */
  protected void outByte(byte byteForOut){
    switch ( outCase ) {
      case 0:
        outputBuffer[outputBufferCount] = byteForOut;
        outputBufferCount++;
        if (outputBufferCount == streamBufferSize)
          flushOutBuffer();
        break;
      case 1:
        BOutStreamA.write(byteForOut);
        break;
      case 2:
        BOutStreamB.write(byteForOut);
        break;
    }
  }

  /**
   * Out Short value to the output stream.
   * @param shortForOut value for output.
   */
  protected void outShort(short shortForOut){
    outByte((byte)((shortForOut >>> 0) & (short)0x00FF));
    outByte((byte)((shortForOut >>> 8) & (short)0x00FF));
  }

  /**
   * Open input stream.
   * You do not invoke this function directly.
   * Method execute invoike this function.
   * @return open success.
   */
  public boolean openInputStream(){
    boolean result;
    result = true;
    if ( this.blankInputStream == null ){
      try{
        this.blankInputStream = new FileInputStream(inputBlank);
      } catch(Exception e){
        this.blankInputStream = null;
        result = false;
      }
    }
    return result;
  };

  /**
   * Close input stream.
   * You do not invoke this function directly.
   * Method execute invoike this function.
   * @return close success.
   */
  public void closeInputStream(){
    try{
      this.blankInputStream.close();
    } catch(Exception e){
    }
    this.blankInputStream = null;
  };

  /**
   * Open output stream.
   * You do not invoke this function directly.
   * Method execute invoike this function.
   * @return open success.
   */
  public boolean openOutputStream(){
    boolean result;
    result = true;
    if ( this.reportOutStream == null ){
      try{
        this.reportOutStream = new FileOutputStream(outputReport);
        outputBufferCount = 0;
      } catch(Exception e){
        this.reportOutStream = null;
        result = false;
      }
    }
    BOutStreamA = new BOutStream(4096);
    BOutStreamB = new BOutStream(8192);
    return result;
  };

  /**
   * Close output stream.
   * You do not invoke this function directly.
   * Method execute invoike this function.
   * @return close success.
   */
  public void closeOutputStream(){
    closeOutBuffer();
    try{
      this.reportOutStream.close();
    } catch(Exception e){
    }
    this.reportOutStream = null;
    BOutStreamA = null;
    BOutStreamB = null;
  };

  /**
   * Virtual auxiliary method.
   */
  protected void outSpecial() {
  }

  /**
   * Abstract method for override in inheritance classes
   * for parse file header section. For different formats
   * (RTF, HTML, XLS4) this methos should do different work.
   * @return success of operation.
   */
  protected abstract boolean prepareFileHeader();
  /**
   * Abstract method for override in inheritance classes
   * for parse file footer section. For different formats
   * (RTF, HTML, XLS4) this methos should do different work.
   * @return success of operation.
   */
  protected abstract boolean prepareFileFooter();
  /**
   * Abstract method for override in inheritance classes
   * for parse report section. For different formats
   * (RTF, HTML, XLS4) this methos should do different work.
   * @param SectionName section name for prepare.
   * @param SectionNum section number for prepare.
   * @return section number.
   */
  protected abstract int prepareSectionInternal(String SectionName, int SectionNum);
  /**
   * Abstract method for override in inheritance classes
   * for output report section. For different formats
   * (RTF, HTML, XLS4) this methos should do different work.
   * @param SectionNum section number for out.
   * @param defaultOut default logic value means outed it by
   * default or not.
   */
  protected abstract void outSectionInternal(int SectionNum, boolean defaultOut);

}