package VKRep;

import java.io.*;

/**
 * <p>Title: </p>
 * <p>Description: This container combine horizontal and vertical
 * report processing in one process. In a result is cross-tab report
 * generator.</p>
 * <p>The output stream from horizontal report use as blank for
 * vertical report. This is the main idea.</p>
 * <p>For XLS4 cross-tab report you should make one blank in XLS4
 * format with horizontal and virtical sections definition within.</p>
 * <p>Then make to objects TXLS4ReportHorizontal and TXLS4ReportVertical.</p>
 * <p>For TXLS4ReportHorizontal object set as input blank your design blank.
 * For TXLS4ReportVertical object set as output report desired file.
 * Set RptObjH and RptObjV property of TVKRptContainerCrossTab object, indicate
 * horisontal and vertical listeners and call execute method.</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Vlad Karpov
 * @version 1.0
 */

public class TVKRptContainerCrossTab extends TVKRptContainerAbstract {

  /**
   * The object of main horizontal report generator.
   * It may be any TReportHorizontal descendant class object
   * (TXLS4ReportHorizontal ( :-) it only one horizontal
   * class now ), ....)
   */
  protected TReportHorizontal rptObjH;

  /**
   * The object of main vertical report generator.
   * It may be TReportVertical descendant class object
   * (TXLS4ReportVertical, ....). You use only compatibility
   * by format objects. It meens that you not use
   * TXLS4ReportHorizontal as horizontal report and TWordReport
   * as a vertical report.
   * Now there is only one TReportHorizontal class - TXLS4ReportHorizontal,
   * so you can do only XLS4 cross-tab report. Use for this
   * TXLS4ReportHorizontal and TXLS4ReportVertical objects.
   */
  protected TReportVertical rptObjV;

  /**
   * The horizontal Listener object. That is exactly "only one" code to
   * fill data of rptObj "input blank".
   */
  protected VKRepListener listenerH;

  /**
   * The vertical Listener object. That is exactly "only one" code to
   * fill data of rptObj "input blank".
   */
  protected VKRepListener listenerV;

  /**
   * Intermediate stream for hold stream from horizontal process
   * to vertical process.
   */
  protected BOutStream bOutStream;

  /**
   * Only constructor.
   */
  public TVKRptContainerCrossTab() {
    rptObjH = null;
    listenerH = null;
    rptObjV = null;
    listenerV = null;
    bOutStream = new BOutStream(4096);
  }

  /**
   * The read property method of rptObjV variable.
   * @return TReportVertical object is setting now.
   */
  public TReportVertical getRptObjV() {
    return rptObjV;
  }

  /**
   * The write property method of rptObj variable.
   * @param value TReportVertical object to setting now.
   */
  public void setRptObjV(TReportVertical value){
    rptObjV = value;
  }

  /**
   * The read property method of rptObj variable.
   * @return TReportHorizontal object is setting now.
   */
  public TReportHorizontal getRptObjH() {
    return rptObjH;
  }

  /**
   * The write property method of rptObj variable.
   * @param value TReportHorizontal object to setting now.
   */
  public void setRptObjH(TReportHorizontal value){
    rptObjH = value;
  }

  /**
   * The read property method of vertical listener variable.
   * @return VKRepListener object is setting now.
   */
  public VKRepListener getListenerV() {
    return listenerV;
  }

  /**
   * The write property method of vertical listener variable.
   * @param value VKRepListener object to setting now.
   */
  public void setListenerV(VKRepListener value) {
    listenerV = value;
  }

  /**
   * The read property method of horizontal listener variable.
   * @return VKRepListener object is setting now.
   */
  public VKRepListener getListenerH() {
    return listenerH;
  }

  /**
   * The write property method of horizontal listener variable.
   * @param value VKRepListener object to setting now.
   */
  public void setListenerH(VKRepListener value) {
    listenerH = value;
  }

  /**
   * This method is start cross-tab generate report.
   */
  public void execute(){
    if ( ( (rptObjV != null) && (listenerV != null) ) &&
         ( (rptObjH != null) && (listenerH != null) ) &&
         ( rptObjV instanceof TReportVertical ) &&
         ( rptObjH instanceof TReportHorizontal )
        ) {
      VKRepListener originalListenerV = rptObjV.getListener();
      VKRepListener originalListenerH = rptObjH.getListener();
      OutputStream originalOutputStream = rptObjH.getReportOutStream();
      try {
        rptObjH.setListener(listenerH);
        rptObjH.setReportOutStream(bOutStream);
        bOutStream.reset();
        rptObjH.execute();
        ByteArrayInputStream instream = new ByteArrayInputStream(bOutStream.getBuffer());
        rptObjV.setListener(listenerV);
        InputStream originalInputStream = rptObjV.getBlankInputStream();
        try {
          rptObjV.setBlankInputStream(instream);
          rptObjV.execute();
        } finally {
          rptObjV.setBlankInputStream(originalInputStream);
        }
      } finally {
        rptObjV.setListener(originalListenerV);
        rptObjH.setListener(originalListenerH);
        rptObjH.setReportOutStream(originalOutputStream);
      }
    }
  };

}