package VKRep;

import java.io.*;

/**
 * <p>Title: Reporter to RTF, HTML, TXT.</p>
 * <p>Description: Main class for all vertical text based
 * reports (TXT, RTF, HTML). </p>
 * <h1>Build report processing.</h1>
 * <p>First of all you should make blank for you report with format what
 * you want. Class TWordReport allowed to do reports in text based
 * formats such as TXT, RTF, HTML. So if you want to get TXT file
 * with you data you should make TXT format blank, if you want
 * to get RTF file you should make RTF blank, and HTML blank if
 * you want HTML.</p>
 * <p>The blank in any formats may consists of report sections.
 * There are 11 standard sections and 2 special sections.
 * In addition you can define your own report sections.</p>
 * <p>11 standard sections:</p>
 *   <ul>
 *   <li> REPORT_HEADER 	= 1;
 *   <li> GROUP_HEADER		= 2;
 *   <li> SUBGROUP_HEADER	= 3;
 *   <li> SUBSUBGROUP_HEADER	= 4;
 *   <li> SUPERSUBGROUP_HEADER	= 5;
 *   <li> BODY			= 6;
 *   <li> SUPERSUBGROUP_FOOTER	= 7;
 *   <li> SUBSUBGROUP_FOOTER	= 8;
 *   <li> SUBGROUP_FOOTER	= 9;
 *   <li> GROUP_FOOTER		= 10;
 *   <li> REPORT_FOOTER	        = 11;
 *   </ul>
 * <p>2 special sections:</p>
 *   <ul>
 *   <li> FILE_HEADER  	        = 0;
 *   <li> FILE_FOOTER  	        = 12;
 *   </ul>
 * <p>All sections have a name and a number. All standard
 * sections have a predefined name and numbers from 0 to 12.
 * When you define your own section you give name this section
 * and when you register this section in reporter
 * ( method int prepareSection(String SectionName) ) you supplied
 * number of section (returned from prepareSection).
 * You can invoke prepareSection only in doSectionPrepare or in
 * prepareSection event.</p>
 * <p>The section definition looks differ in differ formats.
 * For example in RTF format REPORT_HEADER section
 * definition looks like:</p>
 * <pre>
 * #BEGIN SECTION 'REPORT HEADER'
 *    ...
 * #END SECTION 'REPORT HEADER'
 * </pre>
 * <p>The same in HTML format:</p>
 * <pre>
 * &#60;!-- { #BEGIN SECTION \lquote REPORT HEADER\rquote } --&#62;
 *    ...
 * &#60;!-- { #END SECTION \lquote REPORT HEADER\rquote } --&#62;
 * </pre>
 * <p>And for TXT format:</p>
 * <pre>
 * { #BEGIN SECTION \lquote REPORT HEADER\rquote }
 *    ...
 * { #END SECTION \lquote REPORT HEADER\rquote }
 * </pre>
 *<p>Two special sections (FILE_HEADER and FILE_FOOTER) is needed
 * for define work area in blank where dispose all other sections.</p>
 *<p>The FILE_HEADER section is part of blank from
 * begining of the blank file to the definition:</p>
 * <pre>
 * #BEGIN VKREPORT
 * </pre>
 * <p>And the FILE_FOOTER section is part of blank from
 * <pre>
 * #END VKREPORT
 * </pre>
 * to the end of blank file.</p>
 * <p>In that way, if we are view (for example) RTF blank in
 * low level, in text editor as a text file we can see the follow:</p>
 * <pre>
 *
 * begin of blank file
 *       |       | {\rtf1\ ...
 *       |file   | ...
 *       |header | {... #BEGIN VKREPORT ...}
 *       |
 *       |         ...
 *       |
 *       |       | {... #BEGIN SECTION \lquote REPORT HEADER\rquote ... }
 *       |report |   ...
 *       |header | {... #END SECTION \lquote REPORT HEADER\rquote ... }
 *       |
 *       |       | {... #BEGIN SECTION \lquote BODY\rquote ... }
 * blank |body   |   ...
 * file  |       | {... #END SECTION \lquote BODY\rquote ... }
 *       |
 *       |       | {... #BEGIN SECTION \lquote REPORT FOOTER\rquote ... }
 *       |report |   ...
 *       |footer | {... #END SECTION \lquote REPORT FOOTER\rquote ... }
 *       |
 *       |         ...
 *       |
 *       |file   | {... #END VKREPORT ... }
 *       |footer | }
 * end of blank file
 *
 * </pre>
 * <p>Properly speaking, all reporter to do is only copy sections
 * from input blank to the output report according the following
 * schema:</p>
 * <pre>
 *
 *   copy FILE_HEADER (always)
 *   copy REPORT_HEADER (try always)
 *     doMoveTop  //move your data source to the top
 *     loop ( not end of your data source )
 *       copy GROUP_HEADER
 *       copy SUBGROUP_HEADER
 *       copy SUBSUBGROUP_HEADER
 *       copy SUPERSUBGROUP_HEADER
 *       copy BODY (try always)
 *       copy SUPERSUBGROUP_FOOTER
 *       copy SUBSUBGROUP_FOOTER
 *       copy SUBGROUP_FOOTER
 *       copy GROUP_FOOTER
 *       doMoveNext //move your data source to the next row
 *     end loop
 *   copy REPORT_FOOTER (try always)
 *   copy FILE_FOOTER (always)
 *
 * </pre>
 * <p>Process of copy any section is accompany following events:</p>
 * <pre>
 *
 *   doSectionBegin            //notify event
 *   doSectionCheck            //qury event to ask copy section or not
 *   if SectionCheck == true then
 *     copy content of section with query data from your data source
 *     doSectionComplete       //notify event
 *   end if
 *   doSectionEnd              //notify event
 *
 * </pre>
 * <p>As you see, you can define in program youself when to copy
 * section exactly (group sections should copy only for definite
 * condition).</p>
 * <p>The data for report is supplied by mark. There are two types
 * of marks: mark by name and mark by number.</p>
 * <p>Mark by name looks like:</p>
 * <pre>
 *   ...
 *   ~field_name1~    ~field_name2~  ....
 *   ...
 * </pre>
 * <p>Mark by number looks like:</p>
 * <pre>
 *   ...
 *   &#64;0001    &#64;0002  ....
 *   ...
 * </pre>
 * <p>When reporter copy section and determinate data request
 * mark (by name or by numbar) it fire event for delivery data.
 * Your programm should handle this event and return data
 * corresponding to mark.</p>
 *
 * <p>The navigation within data source realize throw two events:
 * onMoveTop and onMoveNext.</p>
 *
 *   <h2>Create blank for you report.</h2>
 * <p>Create blank process include 2 steps in general. First -
 * edit blank in high level tools, such as WordPard, WinWord,
 * FronPage and so on.</p>
 * <p>The second step is needed for correct defects in blank
 * after high level tools and it happens in low level tools
 * like a simple text editor.</p>
 *
 *     <h3>Blank for RTF format.</h3>
 *       <h4>High level.</h4>
 * <p>Open WinWord. Define work area to type #BEGIN VKREPORT at begin
 * and #END VKREPORT at end.</p>
 * <p>I usually to set of blue color for #BEGIN VKREPORT and #END VKREPORT.</p>
 * <p>Between #BEGIN VKREPORT and #END VKREPORT define any report
 * sections with data request mark within section. Any section start
 * with #BEGIN SECTION 'SECTION NAME' and end of #END SECTION 'SECTION NAME'.
 * Tag #BEGIN SECTION I usually set to green color and tag #END SECTION
 * set to gray color. The color set is needed for correct compound
 * VK Report tag within a RTF tag.</p>
 * <p>You can apply any formating text, paragraph. Use tables, lists and any
 * ather features RTF format - these all available for you reports.</p>
 *       <h4>Low level.</h4>
 * <p>This step is needed because WinWord may mix VK Report control tags within
 * RTF control tags and VK Reports can not find these.
 * Perceive this step like instraction:
 * <ul>
 *   <li> Make bakup copy of your blank.
 *   <li> Load blank any text editor (I like Multy-Edit).
 *   <li> Find by standard find dialog in editor all entries of symbol '#'.
 *   <li> Analyse these entries. If it tags for VK Reports, VK Reports must
 *        be able to find these one. It meens, that if you see something like
 *<pre>
 *{\v\cf10\lang1033\langfe1033\langnp1033 #BEGIN SECTION \lquote REPORT {\lang1033 HEADER}\rquote
 *\par }
 *</pre>
 *        the VK Reports not find key string - '#BEGIN SECTION \lquote REPORT HEADER}\rquote'.
 *        It's bad.
 *        You should delete unnecessary symbols by hands. After modifications the RTF
 *        tag must looks like:
 *<pre>
 *{\v\cf10\lang1033\langfe1033\langnp1033 #BEGIN SECTION \lquote REPORT HEADER\rquote
 *\par }
 *</pre>
 *   <li> Find by standard find dialog in editor all entries of symbol '@'.
 *   <li> Analyse these entries. The expressions like @0{\b 0}01 inadmissible.
 *   <li> If you use charset another than American (ISO... I don't know...) you
 *        should change default charset in RTF header. Change \fcharset0 to
 *        \fcharsetN where N is number of you charset table (204 for Russian).
 *        Note... you can out data to RTF by data request mark in UNICUDE, and you
 *        don't need change charset table.
 *   <li> Save blank.
 *   <li> Try to open blank with WinWord. If all right, than the process is finished.
 *        If you can not open blank with WinWord - you have bakup copy...
 *   </ul>
 * </p>
 *     <h3>Blank for HTML format.</h3>
 *       <h4>High level.</h4>
 * In any high level tools (Macromedia Dreamweaver MX, MS FronPage, and so on ...) design your
 * report blank. You should segregate report sections from each other and
 * keep it mind when you design report blank. You can use any HTML features
 * in your blank without limitations.
 *       <h4>Low level.</h4>
 * In this step you should insert VK Reports tags in your HTML blank by hands in
 * any simple text editor. The VK Reports tags for HTML is the same for RTF and
 * looks like:
 * <pre>
 *
 * < begin of your HTML report blank >
 *     ...
 * < !-- { #BEGIN VKREPORT } -- >
 * < !-- { #BEGIN SECTION \lquote REPORT HEADER\rquote } -- >
 *     ... @0001  ... @0002 ...
 * < !-- { #END SECTION \lquote REPORT HEADER\rquote } -- >
 *     ...
 * < other sections >
 *     ...
 * < !-- { #BEGIN SECTION \lquote BODY\rquote } -- >
 *     ... @0001  ... @0002 ...
 * < !-- { #END SECTION \lquote BODY\rquote } -- >
 *     ...
 * < other sections >
 *     ...
 * < !-- { #BEGIN SECTION \lquote REPORT FOOTER\rquote }  -- >
 *     ... @0001  ... @0002 ...
 * < !-- { #END SECTION \lquote REPORT FOOTER\rquote }  -- >
 * < !-- { #END VKREPORT } -- >
 *     ...
 * < end of your HTML report blank >
 *
 * </pre>
 *
 * As you see VK Reports tags design as HTML remarks with RTF tags inside
 * for compatibility between these formats.
 *
 *     <h3>Blank for TXT format.</h3>
 *
 * To make TXT report blank you should do the same that you do for HTML.
 * (It clear - HTML is the text...). But VK Reports tags is not be HTML remarks.
 * It looks like:
 * <pre>
 *
 * < begin of your TXT report blank >
 *     ...
 * { #BEGIN VKREPORT }
 * { #BEGIN SECTION \lquote REPORT HEADER\rquote }
 *     ... @0001  ... @0002 ...
 * { #END SECTION \lquote REPORT HEADER\rquote }
 *     ...
 * < other sections >
 *     ...
 * { #BEGIN SECTION \lquote BODY\rquote }
 *     ... @0001  ... @0002 ...
 * { #END SECTION \lquote BODY\rquote }
 *     ...
 * < other sections >
 *     ...
 * { #BEGIN SECTION \lquote REPORT FOOTER\rquote }
 *     ... @0001  ... @0002 ...
 * { #END SECTION \lquote REPORT FOOTER\rquote }
 * { #END VKREPORT }
 *     ...
 * < end of your HTML report blank >
 *
 * </pre>
 *
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Vlad Karpov
 * @version 1.0
 * @see <a href = "TReportAbstract.html">TReportAbstract</a>
 */

public class TWordReport extends TReportVertical {

  public TWordReport() {
  }

  protected void outSectionInternal(int SectionNum, boolean defaultOut){
    int i, j, k;
    int secOff;
    byte a0, a1, a2, a3;
    int dataIndex;
    if ( sections[SectionNum].sectionUsed ){
      this.rptEvent.setSectionName(sections[SectionNum].sectionName);
      this.rptEvent.sectionNum = SectionNum;
      doSectionBegin(this.rptEvent);
      try{
        this.rptEvent.check = defaultOut;
        doSectionCheck(this.rptEvent);
        if ( this.rptEvent.check ){
          try {
            i = 0;
            while ( i < sections[SectionNum].sectionLength ){
              secOff = sections[SectionNum].sectionStartPosition + i;
              //** Request data by number (@0001, @0002 ...) */
              if ( inputBuffer[secOff] == charByNum ){
                a0 = inputBuffer[secOff + 1];
                a1 = inputBuffer[secOff + 2];
                a2 = inputBuffer[secOff + 3];
                a3 = inputBuffer[secOff + 4];
                if ( ((0x30 <= a0) && (a0 <= 0x39)) &&
                     ((0x30 <= a1) && (a1 <= 0x39)) &&
                     ((0x30 <= a2) && (a2 <= 0x39)) &&
                     ((0x30 <= a3) && (a3 <= 0x39)) ) {
                  dataIndex = (a0 - 0x30) * 1000 + (a1 - 0x30) * 100 + (a2 - 0x30) * 10 + (a3 - 0x30);
                  if ( dataIndex != 0 ) {  //Begin with 1  !!!!
                    this.rptEvent.dataIndex = dataIndex;
                    this.rptEvent.outDataLen = 0;
                    /** In this call you should fill outData and outDataLen fields in rptEvent object. */
                    doRequestByNum(this.rptEvent);
                    /** */
                    j = 0;
                    while ( j < this.rptEvent.outDataLen ){
                      outByte((byte)this.rptEvent.outData[j]);
                      j++;
                    }
                    i+=5;
                  } else {
                    outByte(inputBuffer[secOff]);
                    i++;
                  }
                } else {
                  outByte(inputBuffer[secOff]);
                  i++;
                }
              } else {
                //** Request data by name (~name1~, ~name2~ ...) */
                if ( inputBuffer[secOff] == charByName ){
                  //** Find next charByName*/
                  loop:{
                    for (k = 1; k < MAX_LABEL_NAME_LEGTH; k++) {
                      if ( inputBuffer[secOff + k] == charByName ){
                        break loop;
                      }
                    }
                  }
                  /** if find next charByName*/
                  if ( k < MAX_LABEL_NAME_LEGTH ){
                    this.rptEvent.dataName = new String(inputBuffer, secOff + 1, k - 1);
                    this.rptEvent.outDataLen = 0;
                    /** In this call you should fill outData and outDataLen fields in rptEvent object. */
                    doRequestByName(this.rptEvent);
                    /** */
                    j = 0;
                    while ( j < this.rptEvent.outDataLen ){
                      outByte((byte)this.rptEvent.outData[j]);
                      j++;
                    }
                    i += (k + 1);
                  } else {
                    outByte(inputBuffer[secOff]);
                    i++;
                  }
                } else {
                  outByte(inputBuffer[secOff]);
                  i++;
                }
              }
            }
          } finally {
            doSectionComplete(this.rptEvent);
          }
        }
      } finally {
        doSectionEnd(this.rptEvent);
      }
    }
  }

  protected boolean prepareFileHeader() {
    boolean result = false;
    int i = findTag("#BEGIN VKREPORT");
    if ( i >= 0 ) {
      while ( ( i >= 0 ) && ( inputBuffer[i] != RTF_SECTION_BEGIN_BREAKED ) ) i--;
      if ( i >= 0 ){
        sections[FILE_HEADER].sectionName = "FILE HEADER";
        sections[FILE_HEADER].sectionNum = FILE_HEADER;
        sections[FILE_HEADER].sectionStartPosition = 0;
        sections[FILE_HEADER].sectionLength = i;
        sections[FILE_HEADER].sectionUsed = true;
        result = true;
      } else {
        this.rptEvent.errorMessage = "There is no '{' before '#BEGIN VKREPORT'.";
        doReportError(this.rptEvent);
      }
    } else {
      this.rptEvent.errorMessage = "There is no tag '#BEGIN VKREPORT' in input blank.";
      doReportError(this.rptEvent);
    }
    return result;
  }

  protected boolean prepareFileFooter() {
    boolean result = false;
    int i = findTag("#END VKREPORT");
    if ( i > 0 ) {
      while ( ( i < realInputBlankLength ) && ( inputBuffer[i] != RTF_SECTION_END_BREAKED ) ) i++;
      if ( i < realInputBlankLength ){
        sections[FILE_FOOTER].sectionName = "FILE FOOTER";
        sections[FILE_FOOTER].sectionNum = FILE_FOOTER;
        sections[FILE_FOOTER].sectionStartPosition = i + 1;
        sections[FILE_FOOTER].sectionLength = realInputBlankLength - i - 1;
        sections[FILE_FOOTER].sectionUsed = true;
        result = true;
      } else {
        this.rptEvent.errorMessage = "There is no '}' after '#END VKREPORT'.";
        doReportError(this.rptEvent);
      }
    } else {
      this.rptEvent.errorMessage = "There is no tag '#END VKREPORT' in input blank.";
      doReportError(this.rptEvent);
    }
    return result;
  }

  protected int prepareSectionInternal(String SectionName, int SectionNum){
    String beginString = "#BEGIN SECTION \\lquote " + SectionName  + "\\rquote";
    String endString = "#END SECTION \\lquote " + SectionName + "\\rquote";
    int i = findTag(beginString);
    if ( i > 0 ) {
      int j = findTag(endString);
      if ( j > 0 ) {
        while ( ( i < realInputBlankLength ) && ( inputBuffer[i] != RTF_SECTION_END_BREAKED ) ) i++;
        if ( i < realInputBlankLength ){
          while ( (j >= 0) && (inputBuffer[j] != RTF_SECTION_BEGIN_BREAKED)) j--;
          if (j >= 0) {
            if ( ( j - (i + 1) ) > 0 ) {
              if (SectionNum == -1) {
                SectionNum = lastDynSection;
                lastDynSection++;
              }
              sections[SectionNum].sectionName = SectionName;
              sections[SectionNum].sectionNum = SectionNum;
              sections[SectionNum].sectionStartPosition = i + 1;
              sections[SectionNum].sectionLength = j - sections[SectionNum].sectionStartPosition;
              sections[SectionNum].sectionUsed = true;
            } else {
              this.rptEvent.errorMessage = "Length prepared section '" + SectionName + "' less zerro.";
              doReportError(this.rptEvent);
            }
          } else {
            this.rptEvent.errorMessage = "There is no '{' before section '" + SectionName + "'.";
            doReportError(this.rptEvent);
          }
        } else {
          this.rptEvent.errorMessage = "There is no '}' after section '" + SectionName + "'.";
          doReportError(this.rptEvent);
        }
      } else {
        this.rptEvent.errorMessage = "There is no END tag for section '" + SectionName  + "' in input blank.";
        doReportError(this.rptEvent);
      }
    } else {
      if (SectionNum == -1) {
        this.rptEvent.errorMessage = "There is no BEGIN tag for section '" + SectionName + "' in input blank.";
        doReportError(this.rptEvent);
      }
    }
    return SectionNum;
  }

  private int findTag(String tag) {
    byte[] btag = tag.getBytes();
    int tagLength = btag.length;
    int[] Lambda = new int[256];
    int j;
    int i;
    int k;
    int result = -1;

    for (i=0; i<256; i++) Lambda[i] = 0;
    for (i=0; i<tagLength; i++) Lambda[btag[i] + 128] = i + 1;

    i = 0;
    mainLoop: {
      while (i < realInputBlankLength) {
        search: {
          j = tagLength - 1;
          while (j >= 0) {
            if (inputBuffer[i + j] == btag[j]) {
              j--;
            } else {
              k = j - Lambda[inputBuffer[i + j] + 128] + 1;
              if ( k  > 0 ) {
                i += k;
              } else {
                i++;
              }
              break search;
            }
          }
        }
        if ( j == -1 ) {
          result = i;
          break mainLoop;
        }
      }
    }
    return result;
  }

}