package VKRep;

import java.io.*;
import java.util.*;

/**
 * <p>Title: </p>
 * <p>Description: Main class for all horizontal XLS4 based
 * reports.</p>
 *
 * <h1>Build report processing.</h1>
 * <p>First of all you should make blank for you report in
 * XLS4 format. The simplest way to do this - open MS Excel and
 * design required blank. Save your blank in XLS format of 4
 * virsion.</p>
 * <p>The blank may consists of report sections.
 * There are 11 standard sections and 2 special sections.
 * In addition you can define your own report sections.</p>
 * <p>11 standard sections for XLS4:</p>
 *   <ul>
 *   <li> REPORT_HEADER 	= 1;
 *   <li> GROUP_HEADER		= 2;
 *   <li> SUBGROUP_HEADER	= 3;
 *   <li> SUBSUBGROUP_HEADER	= 4;
 *   <li> SUPERSUBGROUP_HEADER	= 5;
 *   <li> BODY			= 6;
 *   <li> SUPERSUBGROUP_FOOTER	= 7;
 *   <li> SUBSUBGROUP_FOOTER	= 8;
 *   <li> SUBGROUP_FOOTER	= 9;
 *   <li> GROUP_FOOTER		= 10;
 *   <li> REPORT_FOOTER	        = 11;
 *   </ul>
 * <p>2 special sections:</p>
 *   <ul>
 *   <li> FILE_HEADER  	        = 0;
 *   <li> FILE_FOOTER  	        = 12;
 *   </ul>
 * <p>All sections have a name and a number. All standard
 * sections have a predefined name and numbers from 0 to 12.
 * When you define your own section you give name this section
 * and when you register this section in reporter
 * ( method int prepareSection(String SectionName) ) you supplied
 * number of section (returned from prepareSection).
 * You can invoke prepareSection only in doSectionPrepare or in
 * prepareSection event.</p>
 * <p>The section definition in XLS4 format - it is simply named
 * range of cells. For horizontal XLS4 report the best way to give
 * names for range columns of cells. The names of sections is standards
 * names without blanks and H_ prefix:</p>
 *   <ul>
 *   <li> H_REPORTHEADER 	= 1;
 *   <li> H_GROUPHEADER		= 2;
 *   <li> H_SUBGROUPHEADER	= 3;
 *   <li> H_SUBSUBGROUPHEADER	= 4;
 *   <li> H_SUPERSUBGROUPHEADER	= 5;
 *   <li> H_BODY		= 6;
 *   <li> H_SUPERSUBGROUPFOOTER	= 7;
 *   <li> H_SUBSUBGROUPFOOTER	= 8;
 *   <li> H_SUBGROUPFOOTER	= 9;
 *   <li> H_GROUPFOOTER		= 10;
 *   <li> H_REPORTFOOTER	= 11;
 *   </ul>
 * <p>You shouldn't define H_FILE_HEADER and H_FILE_FOOTER. Your own
 * sections names shouldn't have blanks within.</p>
 * <p>Properly speaking, all reporter to do is only copy BIFF recors
 * from input blank to the output report according the following
 * schema:</p>
 * <pre>
 *
 *   copy FILE_HEADER (always)
 *   copy REPORT_HEADER (try always)
 *     doMoveTop  //move your data source to the top
 *     loop ( not end of your data source )
 *       copy GROUP_HEADER
 *       copy SUBGROUP_HEADER
 *       copy SUBSUBGROUP_HEADER
 *       copy SUPERSUBGROUP_HEADER
 *       copy BODY (try always)
 *       copy SUPERSUBGROUP_FOOTER
 *       copy SUBSUBGROUP_FOOTER
 *       copy SUBGROUP_FOOTER
 *       copy GROUP_FOOTER
 *       doMoveNext //move your data source to the next row
 *     end loop
 *   copy REPORT_FOOTER (try always)
 *   copy FILE_FOOTER (always)
 *
 * </pre>
 * <p>Process of copy any section is accompany following events:</p>
 * <pre>
 *
 *   doSectionBegin            //notify event
 *   doSectionCheck            //qury event to ask copy section or not
 *   if SectionCheck == true then
 *     copy content of section with query data from your data source
 *     doSectionComplete       //notify event
 *   end if
 *   doSectionEnd              //notify event
 *
 * </pre>
 * <p>As you see, you can define in program youself when to copy
 * section exactly (group sections should copy only for definite
 * condition).</p>
 * <p>The data for report is supplied by mark and by consists of cell.
 * There are two types of marks: mark by name and mark by number.
 * First, in OnDataRequest event you can define data for concrete cell.
 * Then (if you not handle event OnDataRequest) try replace all mark
 * (by name and by nember) in OnRequestByName and
 * OnRequestByNum events.</p>
 * <p>Mark by name looks like:</p>
 * <pre>
 *   ...
 *   ~field_name1~    ~field_name2~  ....
 *   ...
 * </pre>
 * <p>Mark by number looks like:</p>
 * <pre>
 *   ...
 *   &#64;0001    &#64;0002  ....
 *   ...
 * </pre>
 * <p>When reporter copy section it copy one cell by cell.
 * When reporter copy cell in output stream (file) the OnDataRequest
 * event is fire where you can define new content of cell. If you not define data
 * in OnDataRequest event (rptEvent.check = false) reporter try to find
 * marks (by name or by number) in a cell. The reporter determinate data request
 * mark it fire event for delivery data.
 * Your programm should handle this event and return data
 * corresponding to mark.</p>
 * <p>Note that XLS4 format support only 255 charster length in cell.</p>
 *
 * <p>The navigation within data source realize throw two events:
 * onMoveTop and onMoveNext.</p>
 *
 *   <h2>Create blank for you XLS4 report.</h2>
 * <p>Create blank process include 1 steps in general !!!!
 * Open MS Excel and make your blank with sections what you need
 * and numerate cells for reflect on OnDataRequest event. Save
 * blank as XLS 4 format. That is all...</p>
 *
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 */

public class TXLS4ReportHorizontal extends TReportHorizontal {

  public TXLS4ReportHorizontal() {
    biffBody = new Vector();
    offsetXLSInput = 0;
    lastHandle = 0;
    charByNum = 38; //&
  }

  private Vector biffBody;
  private int lastHandle = 0;
  private int currCol = 0;

  protected void clearSections(){
    super.clearSections();
    offsetXLSInput = 0;
    lastHandle = 0;
    currCol = 0;
    biffBody.clear();
  }

  protected void clearVars(){
    currCol = 0;
  };

  protected boolean prepareFileHeader() {
    TBiffRec pp;
    TBiffRecHeader r_h = new TBiffRecHeader();
    sections[FILE_HEADER].sectionName = "FILE HEADER";
    sections[FILE_HEADER].sectionNum = FILE_HEADER;
    sections[FILE_HEADER].sectionUsed = true;
    dimentions_loop: {
      while ( true ) {
        r_h.opcode = getShort();
        r_h.length = getShort();
        pp = new TBiffRec();
        pp.head.opcode = r_h.opcode;
        pp.head.length = r_h.length;
        if ( r_h.length > 0 ) {
          pp.data = new byte[r_h.length];
          getBytes(pp.data);
        }
        if ( r_h.opcode == 0x7D ) { //Col
          biffBody.add(pp);
          break dimentions_loop;
        }
        if ( r_h.opcode == 0x200 ) break dimentions_loop;
        sections[FILE_HEADER].contents.add(pp);
      }
    }
    rest_loop: {
      while ( true ) {
        r_h.opcode = getShort();
        r_h.length = getShort();
        pp = new TBiffRec();
        pp.head.opcode = r_h.opcode;
        pp.head.length = r_h.length;
        if ( r_h.length > 0 ) {
          pp.data = new byte[r_h.length];
          getBytes(pp.data);
        }
        biffBody.add(pp);
        if ( r_h.opcode == 0x0A  ) break rest_loop; //EOF
      }
    }
    return true;
  }

  protected boolean prepareFileFooter(){
    int i;
    sections[FILE_FOOTER].sectionName = "FILE FOOTER";
    sections[FILE_FOOTER].sectionNum = FILE_FOOTER;
    sections[FILE_FOOTER].sectionUsed = true;
    for ( i=lastHandle+1; i<biffBody.size(); i++)
      sections[FILE_FOOTER].contents.add(biffBody.elementAt(i));
    return true;
  }

  protected int prepareSectionInternal(String SectionName, int SectionNum){
    TBiffRec pp, qq;
    int i;
    short r, r1 = 0, r2 = 0;
    short c, c1 = 0, c2 = 0;
    short col_b = 0, col_e = 0;
    byte sLen;
    String sName;
    SectionName = removeAllBlanks(SectionName);
    if (SectionNum != -1) {
      SectionName = "H_" + SectionName;
    }
    boolean Found = false;
    for (i = 0; i < sections[FILE_HEADER].contents.size(); i++) {
      pp = (TBiffRec)(sections[FILE_HEADER].contents.elementAt(i));
      if ( pp.head.opcode == 536 ){ //Names definition
        sLen = pp.data[3];
        sName = new String(pp.data, 6, sLen);
        if ( sName.equals(SectionName) ) {
          r1 = getShort(pp.data, pp.head.length - 6);
          r2 = getShort(pp.data, pp.head.length - 4);
          c1 = byteAsShort(pp.data[pp.head.length - 2]);
          c2 = byteAsShort(pp.data[pp.head.length - 1]);
          Found = true;
          break;
        }
      }
    }
    if ( Found ) {
      if (SectionNum == -1) {
        SectionNum = lastDynSection;
        lastDynSection++;
      }
      sections[SectionNum].clear();
      sections[SectionNum].sectionName = SectionName;
      sections[SectionNum].sectionNum = SectionNum;
      sections[SectionNum].sectionUsed = true;
      sections[SectionNum].rowsCount = r2 - r1 + 1;
      sections[SectionNum].colsCount = c2 - c1 + 1;
      for ( i=0; i<biffBody.size(); i++){
        pp = (TBiffRec)biffBody.elementAt(i);
        if (pp.head.opcode == 125) { //Col definition
          col_b = getShort(pp.data, 0);
          col_e = getShort(pp.data, 2);
          if ( ( (c1 <= col_b) && (col_b <= c2)) || ( (c1 <= col_e) && (col_e <= c2))) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.col_b = col_b;
            qq.col_e = col_e;
            qq.cb = c1;
            qq.d = null;
            if (pp.head.length > 0) {
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if (i > lastHandle)
              lastHandle = i;
          }
        }
        if ( pp.head.opcode == 513 ){  //Blank cell
          r = getShort(pp.data, 0);
          c = getShort(pp.data, 2);
          if ( ( (r1 <= r) && (r <= r2) ) && ( (c1 <= c) && (c <= c2) ) ) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.rb = r - r1;
            qq.cb = c - c1;
            qq.r = r;
            qq.c = c;
            qq.d = null;
            if ( pp.head.length > 0 ) {
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if ( i > lastHandle ) lastHandle = i;
          }
        }
        if ( pp.head.opcode == 515 ) { //8 byte float (double)
          r = getShort(pp.data, 0);
          c = getShort(pp.data, 2);
          if ( ( (r1 <= r) && (r <= r2) ) && ( (c1 <= c) && (c <= c2) ) ) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.rb = r - r1;
            qq.cb = c - c1;
            qq.r = r;
            qq.c = c;
            qq.d = new Double(Double.longBitsToDouble(getLong(pp.data, 6)));
            if ( pp.head.length > 0 ){
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if ( i > lastHandle ) lastHandle = i;
          }
        }
        if ( pp.head.opcode == 516 ) { //Label
          r = getShort(pp.data, 0);
          c = getShort(pp.data, 2);
          if ( ( (r1 <= r) && (r <= r2) ) && ( (c1 <= c) && (c <= c2) ) ) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.rb = r - r1;
            qq.cb = c - c1;
            qq.r = r;
            qq.c = c;
            qq.d = new String(pp.data, 8, getShort(pp.data, 6));
            if ( pp.head.length > 0 ){
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if ( i > lastHandle ) lastHandle = i;
          }
        }
        if ( pp.head.opcode == 517 ) { //Bool
          r = getShort(pp.data, 0);
          c = getShort(pp.data, 2);
          if ( ( (r1 <= r) && (r <= r2) ) && ( (c1 <= c) && (c <= c2) ) ) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.rb = r - r1;
            qq.cb = c - c1;
            qq.r = r;
            qq.c = c;
            if ( getShort(pp.data, 6) == 1 ){
              qq.d = new Boolean(true);
            } else {
              qq.d = new Boolean(false);
            }
            if ( pp.head.length > 0 ){
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if ( i > lastHandle ) lastHandle = i;
          }
        }
        if ( pp.head.opcode == 638 ){  //Cool cell - clear ass
          r = getShort(pp.data, 0);
          c = getShort(pp.data, 2);
          if ( ( (r1 <= r) && (r <= r2) ) && ( (c1 <= c) && (c <= c2) ) ) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.rb = r - r1;
            qq.cb = c - c1;
            qq.r = r;
            qq.c = c;
            qq.d = ass2Number(getInt(pp.data, 6));
            if ( pp.head.length > 0 ) {
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if ( i > lastHandle ) lastHandle = i;
          }
        }
        if ( pp.head.opcode == 1030 ){  //Formula cell
          r = getShort(pp.data, 0);
          c = getShort(pp.data, 2);
          if ( ( (r1 <= r) && (r <= r2) ) && ( (c1 <= c) && (c <= c2) ) ) {
            qq = new TBiffRec();
            qq.head.opcode = pp.head.opcode;
            qq.head.length = pp.head.length;
            qq.data = null;
            qq.rb = r - r1;
            qq.cb = c - c1;
            qq.r = r;
            qq.c = c;
            qq.d = null;
            if ( pp.head.length > 0 ) {
              qq.data = new byte[pp.head.length];
              System.arraycopy(pp.data, 0, qq.data, 0, pp.head.length);
            }
            sections[SectionNum].contents.add(qq);
            if ( i > lastHandle ) lastHandle = i;
          }
        }
      }
    } else {
      if (SectionNum == -1) {
        this.rptEvent.errorMessage = "There is no vertical section '" + SectionName + "' in input blank.";
        doReportError(this.rptEvent);
      }
    }
    return SectionNum;
  }

  protected void outSectionInternal(int SectionNum, boolean defaultOut){
    TBiffRec pp;
    Vector list;
    int i, j, k, l;
    short strLen;
    int secOff;
    byte a0, a1, a2, a3;
    int dataIndex;
    if ( SectionNum == 0 )
      outCase = 1;
    else
      outCase = 2;
    if ( sections[SectionNum].sectionUsed ){
      this.rptEvent.setSectionName(sections[SectionNum].sectionName);
      this.rptEvent.sectionNum = SectionNum;
      doSectionBegin(this.rptEvent);
      try{
        this.rptEvent.check = defaultOut;
        doSectionCheck(this.rptEvent);
        if ( this.rptEvent.check ){
          try {
            list = sections[SectionNum].contents;
            for (i=0; i<list.size(); i++) {
              pp = (TBiffRec)list.elementAt(i);
              switch ( pp.head.opcode ){
                case 513:
                case 515:
                case 516:
                case 517:
                case 638:
                case 1030:

                  //*** Call OnDataReqvest event
                  //**********************************************************
                  //You should fill check field in event
                  //handler to determinate will you out data
                  //in this cell or not
                  this.rptEvent.check = false;
                  //Information about absolute and relative
                  //cell coordinates
                  this.rptEvent.rowAbs = pp.r + 1;
                  this.rptEvent.colAbs = currCol + pp.cb + 1;
                  this.rptEvent.rowRel = pp.r + 1;
                  this.rptEvent.colRel = pp.c + 1;
                  //Content of the cell
                  this.rptEvent.dataInput = pp.d;
                  //If you set check field = true in your handler,
                  //you should fill dataOutput field with object
                  //one of this type (Integer, Double, Boolean,
                  //String). (The Date Excel hold as a Double)
                  this.rptEvent.dataOutput = null;
                  //The String much better pass from your
                  //handler throw outData and
                  //outDataLen fields. In this case you should
                  //fill outData, set outDataLen (in Excel 4
                  //string may not be greet then 255 chars) and
                  //set dataOutput = null
                  //If you want to pass NULL in output you should
                  //set check = true, dataOutput = null, outDataLen = 0.
                  //In case check = true, dataOutput != null, outDataLen != 0
                  //VK Reports prefer dataOutput field.
                  this.rptEvent.outDataLen = 0;
                  doDataRequest(this.rptEvent);
                  //**********************************************************
                  if ( this.rptEvent.check ) {
                    outVariant(pp, pp.r, currCol + pp.cb);
                  } else {
                    //If not handle doDataRequest try handle
                    //doRequestByNum and doRequestByName
                    if ( pp.head.opcode == 516 ) {    //Label
                      this.rptEvent.startPrivateData();
                      j = 0;
                      strLen = getShort(pp.data, 6);
                      while ( j < strLen ){
                        secOff = ( 8 + j );
                        //** Request data by number (&0001, &0002 ...) */
                        if ( pp.data[secOff] == charByNum ){
                          a0 = pp.data[secOff + 1];
                          a1 = pp.data[secOff + 2];
                          a2 = pp.data[secOff + 3];
                          a3 = pp.data[secOff + 4];
                          if ( ( (0x30 <= a0) && (a0 <= 0x39)) &&
                              ( (0x30 <= a1) && (a1 <= 0x39)) &&
                              ( (0x30 <= a2) && (a2 <= 0x39)) &&
                              ( (0x30 <= a3) && (a3 <= 0x39))) {
                            dataIndex = (a0 - 0x30) * 1000 + (a1 - 0x30) * 100 + (a2 - 0x30) * 10 + (a3 - 0x30);
                            if (dataIndex != 0) { //Begin with 1  !!!!
                              this.rptEvent.dataIndex = dataIndex;
                              this.rptEvent.outDataLen = 0;
                              /** In this call you should fill outData and outDataLen fields in rptEvent object. */
                              doRequestByNum(this.rptEvent);
                              /** */
                              k = 0;
                              while (k < this.rptEvent.outDataLen) {
                                this.rptEvent.outPrivateByte((byte)this.rptEvent.outData[k]);
                                k++;
                              }
                              j += 5;
                            } else {
                              this.rptEvent.outPrivateByte(pp.data[secOff]);
                              j++;
                            }
                          } else {
                            this.rptEvent.outPrivateByte(pp.data[secOff]);
                            j++;
                          }
                        } else {
                          //** Request data by name (~name1~, ~name2~ ...) */
                          if ( pp.data[secOff] == charByName ){
                            //** Find next charByName*/
                            loop: {
                              for (l = 1; l < MAX_LABEL_NAME_LEGTH; l++) {
                                if (pp.data[secOff + l] == charByName) {
                                  break loop;
                                }
                              }
                            }
                            /** if find next charByName*/
                            if (l < MAX_LABEL_NAME_LEGTH) {
                              this.rptEvent.dataName = new String(pp.data, secOff + 1, l - 1);
                              this.rptEvent.outDataLen = 0;
                              /** In this call you should fill outData and outDataLen fields in rptEvent object. */
                              doRequestByName(this.rptEvent);
                              /** */
                              k = 0;
                              while (k < this.rptEvent.outDataLen) {
                                this.rptEvent.outPrivateByte((byte)this.rptEvent.outData[k]);
                                k++;
                              }
                              j += (l + 1);
                            } else {
                              this.rptEvent.outPrivateByte(pp.data[secOff]);
                              j++;
                            }
                          } else {
                            this.rptEvent.outPrivateByte(pp.data[secOff]);
                            j++;
                          }
                        }
                      }
                      this.rptEvent.stopPrivateData();
                      outVariant(pp, pp.r, currCol + pp.cb);
                    } else {
                      //If not handled with OnRequestByNum or OnRequestByName
                      //copy with col and row
                      outShort(pp.head.opcode);
                      outShort(pp.head.length);
                      setShort(pp.data, 0, (short) pp.r);
                      setShort(pp.data, 2, (short) (currCol + pp.cb));
                      if (pp.head.length > 0)
                        outBytes(pp.data);
                    }
                  }
                  break;
                case 125: //Col
                  //Copy with correct Col
                  outCase = 1;
                  outShort(pp.head.opcode);
                  outShort(pp.head.length);
                  setShort(pp.data, 0, (short) (currCol + pp.col_b - pp.cb));
                  setShort(pp.data, 2, (short) (currCol + pp.col_e - pp.cb));
                  if (pp.head.length > 0)
                    outBytes(pp.data);
                  outCase = 2;
                  break;
                default:
                  //Only copy
                  outShort(pp.head.opcode);
                  outShort(pp.head.length);
                  if ( pp.head.length > 0 ) outBytes(pp.data);
                  break;
              }
            }
            currCol += sections[SectionNum].colsCount;
          } finally {
            doSectionComplete(this.rptEvent);
          }
        }
      } finally {
        doSectionEnd(this.rptEvent);
      }
    }
  }

  protected void outSpecial() {
    outCase = 1;
    outShorts(XLSDimension);
    outCase = 2;
  }

}