package VKRep;

import java.util.*;

/**
 * <p>Title: </p>
 * <p>Description: Event class for VK Reports.</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Vlad Karpov
 * @version 1.0
 */

public class VKRepEvent extends EventObject {

  /**
   * Maximum length section name
   */
  static final int MAX_LEN_SECTION_NAME     = 256;
  /**
   * Maximum length of byte data for report
   */
  static final int MAX_LEN_BYTE_DATA        = 8192;

  /**
   * Constructor VKRepEvent object.
   * @param source owner of object.
   */
  public VKRepEvent(Object source) {
    super(source);
    sectionName = new char[MAX_LEN_SECTION_NAME];
    outData = new byte[MAX_LEN_BYTE_DATA];
    outPrivateData = new byte[MAX_LEN_BYTE_DATA];
  }

  /**
   * Error Message for reportError event.
   */
  public String errorMessage;
  /**
   * End Of File boolean indicator for moveTop and moveNext events.
   */
  public boolean eof;
  /**
   * Record number for moveTop and moveNext events.
   */
  public int recno;
  /**
   * Current section name. Use in sectionBegin, sectionEnd,
   * sectionComplete, sectionCheck, dataRequest, requestByName,
   * requestByNum, sectionPrepare events.
   */
  public char[] sectionName;
  /**
   * Current section number. Use in sectionBegin, sectionEnd,
   * sectionComplete, sectionCheck, dataRequest, requestByName,
   * requestByNum, sectionPrepare events.
   */
  public int sectionNum;
  /**
   * Boolean indicator for control end of data source or necessity
   * of change data in cell or data by mark.
   */
  public boolean check;
  /**
   * Use in requestByNum event for indicate number of mark.
   */
  public int dataIndex;
  /**
   * Use in requestByName event for indicate name of mark.
   */
  public String dataName;
  /**
   * Use in dataRequest event for indicate absolute and relative
   * cell coordinates.
   */
  public int rowAbs, colAbs, rowRel, colRel;
  /**
   * Use in dataRequest event. May consists of Integer, String,
   * Double, Date objects - old content of cell.
   */
  public Object dataInput;
  /**
   * Use in dataRequest event. May consists of Integer, String,
   * Double, Date objects - new content of cell.
   */
  public Object dataOutput;
  /**
   * Use in requestByName, requestByNum events - it is new text
   * in output blank instead of mark.
   */
  public byte[] outData;
  /**
   * The length of outData. Should fill with outData for place
   * outData in output stream (file).
   */
  public int outDataLen;

  /**
   * Intermediate array for store intermediate data.
   */
  private byte[] outPrivateData;
  /**
   * The length of outPrivateData.
   */
  private int outPrivateDataLen;

  /**
   * Start fill of intermediate buffer.
   */
  public void startPrivateData(){
    outPrivateDataLen = 0;
  }

  /**
   * Out byte into intermediate buffer.
   * @param outByte - byte for out.
   */
  public void outPrivateByte(byte outByte){
    outPrivateData[outPrivateDataLen++] = outByte;
  }

  /**
   * Stop fill of intermediate buffer and carry one in the
   * outData variable.
   */
  public void stopPrivateData(){
    check = true;
    dataOutput = null;
    outDataLen = outPrivateDataLen;
    System.arraycopy(outPrivateData, 0, outData, 0, outPrivateDataLen);
    outPrivateDataLen = 0;
  }

  /**
   * Write method of dataIndex property.
   * @param newValue
   */
  public void setDataIndex(int newValue){
    dataIndex = newValue;
  }

  /**
   * Read method of dataIndex property.
   * @return
   */
  public int getDataIndex(){
    return dataIndex;
  }

  /**
   * Read method of check property.
   * @return
   */
  public boolean getCheck(){
    return check;
  }

  /**
   * Write method of check property.
   * @param newValue
   */
  public void setCheck(boolean newValue){
    check = newValue;
  }

  /**
   * Read method of sectionNum property.
   * @return
   */
  public int getSectionNum(){
    return sectionNum;
  }

  /**
   * Write method of sectionNum property.
   * @param newValue
   */
  public void setSectionNum(int newValue){
    sectionNum = newValue;
  }

  /**
   * Read method of sectionName property.
   * @return
   */
  public String getSectionName(){
    return new String(sectionName);
  }

  /**
   * Write method of sectionName property.
   * @param newValue
   */
  public void setSectionName(String newValue){
    int len = newValue.length();
    newValue.getChars(0, len, sectionName, 0);
    sectionName[len] = 0;
  }

  /**
   * Read method of recno property.
   * @return
   */
  public int getRecno(){
    return recno;
  }

  /**
   * Write method of recno property.
   * @param newValue
   */
  public void setRecno(int newValue){
    recno = newValue;
  }

  /**
   * Read method of eof property.
   * @return
   */
  public boolean getEof(){
    return eof;
  }

  /**
   * Write method of eof property.
   * @param newValue
   */
  public void setEof(boolean newValue){
    eof = newValue;
  }

  /**
   * Read method of errorMessage property.
   * @return
   */
  public String getErrorMessage(){
    return errorMessage;
  }

  /**
   * Write method of errorMessage property.
   * @param newValue
   */
  public void setErrorMessage(String errorMessage){
    this.errorMessage = errorMessage;
  }

}