package VKRep;

import java.util.*;

/**
 * <p>Title: </p>
 * <p>Description: Event listener interface for VK Reports.</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Vlad Karpov
 * @version 1.0
 */

public interface VKRepListener extends EventListener {

  /**
   * VK Reports fire this event when report is going to start.
   * @param e VKRepEvent object.
   */
  public void startReport(VKRepEvent e);

  /**
   * VK Reports fire this event when report process has ended.
   * @param e VKRepEvent object.
   */
  public void endReport(VKRepEvent e);

  /**
   * VK Reports fire this event when error raised in report process.
   * e.errorMessage have an error message.
   * @param e VKRepEvent object.
   */
  public void reportError(VKRepEvent e);

  /**
   * When VK Reports required to move data source to top moveTop event is raised.
   * You should move your data source to top and fill e.eof logic variable
   * to show VK Reports is your data source in End Of File or not.
   * <pre>
   * By default:
   *   e.recno = 1;
   *   e.eof = true;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void moveTop(VKRepEvent e);

  /**
   * When VK Reports required to move data source to the next row
   * moveNext event is raised. You should move your data source to next
   * row fill e.eof logic variable to show VK Reports is your data
   * source in End Of File or not.
   * <pre>
   * By default:
   *   e.recno = < current record >;
   *   e.eof = < last settings >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void moveNext(VKRepEvent e);

  /**
   * This is notify event to show your program that section copy
   * process is started.
   * <pre>
   * INput variable set:
   *   e.sectionName = < current section name >;
   *   e.sectionNum = < current section number >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void sectionBegin(VKRepEvent e);

  /**
   * This is notify event to show your program that section copy
   * process is ended.
   * <pre>
   * INput variable set:
   *   e.sectionName = < current section name >;
   *   e.sectionNum = < current section number >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void sectionEnd(VKRepEvent e);

  /**
   * This is notify event to show your program that section copy
   * process is completed. Instead of sectionBegin and sectionEnd this
   * event is fired when the section is realy has been copied.
   * This happen when sectionCheck event for this section is returned from
   * e.check true.
   * <pre>
   * INput variable set:
   *   e.sectionName = < current section name >;
   *   e.sectionNum = < current section number >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void sectionComplete(VKRepEvent e);

  /**
   * After sectionBegin event VK Reports fire sectionCheck event
   * to ask main process about out the current section into the
   * output stream (file) or not. You program should decide out
   * section or not and turn on/off e.check variable. It aplly
   * in most cases for group and subgroup sections.
   * <pre>
   * INput variable set:
   *   e.sectionName = < current section name >;
   *   e.sectionNum = < current section number >;
   * OUTput variable set:
   *   e.check = < true/false >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void sectionCheck(VKRepEvent e);

  /**
   * In cell based reports (XLS4 Vertical, XLS4 Horizontal, ...)
   * event dataRequest is aimed for supply the cell of a new data.
   * <pre>
   * INput variable set:
   *   e.check = < false >;
   *   e.rowAbs = < Absolute row in output report >;
   *   e.colAbs = < Absolute collumn in output report >;
   *   e.rowRel = < Relative row in input blank >;
   *   e.colRel = < Relative collumn in input blank >;
   *   e.dataInput = < Content of the cell >;
   *     dataInput is Object type and may keep Integer, Double,
   *     Boolean, String variable.
   *   e.dataOutput = < null >;
   *     dataOutput is Object type and may keep Integer, Double,
   *     Boolean, String variable.
   *   e.outDataLen = < 0 >;
   *   e.outData = < empty array of char >;
   * OUTput variable set:
   *   e.check = < true/false >;
   *     true/false - do you supply cell of data or not?
   *   e.dataOutput = < New content of the cell >;
   *     dataOutput is Object type and may keep Integer, Double,
   *     Boolean, String variable.
   *     Note:
   *       1) Date in XLS4 is keep like an Double.
   *       2) The String variable march better pass to report
   *          in e.outData and e.outDataLen event variable.
   *       3) XLS4 holds only 255 bytes string.
   *   e.outDataLen = < Length of outData array >;
   *   e.outData = < fill of your data string array of char >;
   * </pre>
   * The String much better pass from your handler throw
   * outData and outDataLen fields. In this case you should fill
   * outData, set outDataLen (in Excel 4 string may not be greet
   * then 255 chars) and set dataOutput = null. If you want to
   * pass NULL in output you should set check = true,
   * dataOutput = null, outDataLen = 0.
   * In case check = true, dataOutput != null, outDataLen != 0
   * VK Reports prefer dataOutput field.
   * If output e.check = false the VK Reports try to apply
   * requestByName and requestByNum for content of cell of blank.
   * @param e VKRepEvent object.
   */
  public void dataRequest(VKRepEvent e);

  /**
   * Event requestByName is aimed for supply the mark by name
   * (~name1~) of a new data.
   * <pre>
   * INput variable set:
   *   e.dataName = < name of mark ("name1" for example) >;
   *   e.outDataLen = < 0 >;
   *   e.outData = < empty array of char >;
   * OUTput variable set:
   *   e.outDataLen = < Length of outData array >;
   *   e.outData = < fill of your data string array of char >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void requestByName(VKRepEvent e);

  /**
   * Event requestByNum is aimed for supply the mark by number
   * (@0001, @0002, ...) of a new data.
   * <pre>
   * INput variable set:
   *   e.dataIndex = < number of mark (1, 2, ... for example) >;
   *   e.outDataLen = < 0 >;
   *   e.outData = < empty array of char >;
   * OUTput variable set:
   *   e.outDataLen = < Length of outData array >;
   *   e.outData = < fill of your data string array of char >;
   * </pre>
   * @param e VKRepEvent object.
   */
  public void requestByNum(VKRepEvent e);

  /**
   * This event need for prepare your own sections.
   * Only in this event you can call
   * public int TReportAbstract.prepareSection(String SectionName)
   * method for determinate your own section with name SectionName.
   * Before call TReportAbstract.prepareSection you should define
   * this section in input report blank.
   * After your section is defined in input blank and prepared in
   * your code you can out this section in any time of building
   * report process by public void TReportAbstract.outSection(int SectionNum)
   * or public void TReportAbstract.outSection(String SectionName) methods.
   * @param e VKRepEvent object.
   * @see <a href="TReportAbstract.html#prepareSection(java.lang.String)">TReportAbstract.prepareSection(String SectionName)</a>
   * @see <a href="TReportAbstract.html#fireSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.fireSectionPrepare(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#doSectionPrepare(VKRep.VKRepEvent)">TReportAbstract.doSectionPrepare(VKRepEvent e)</a>
   * @see <a href="TReportAbstract.html#outSection(java.lang.String)">TReportAbstract.outSection(java.lang.String)</a>
   */
  public void sectionPrepare(VKRepEvent e);

}